package planet.chord;
import java.math.BigInteger;

import planet.commonapi.Id;
import planet.commonapi.exception.InitializationException;
import planet.util.Properties;
import planet.util.Utilities;

public class ChordId extends Id implements java.io.Serializable {

	/* ******************* CHORD SPECIFIC  CONSTANTS **************/
	/**
	 * Chord specific constant: Maximum number of bits for any Id.
	 */
	public static final int MAX_BITS = 160;
	/**
	 * Chord specific constant:  Number of bits saved at any array position.
	 */
	public static final int MIN_BITS = 32;
	/**
	 * Chord specific constant: Maximum value of internal value of Id.
	 */
	public static final int[] MAX_CHORD = {0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff};
	/**
	 * Chord specific constant: The value of one (1) of internal value of Id.
	 */
	public static final int[] ONE_CHORD = {0x1, 0, 0, 0, 0};
	/**
	 * Chord specific constant: The value of two (2) of internal value of Id.
	 */
	public static final int[] TWO_CHORD = {0x2, 0, 0, 0, 0};
	
	/**
	 * Chord specific constant: Each array position contains a maximum 
	 * representation of MODULE bits.
	 */
	protected static final int MODULE         = MIN_BITS;
	/**
	 * Chord specific constant: Maximum size of array that contains entire identification.
	 */
	protected static final int MAX_ARRAY_SIZE = MAX_BITS/MODULE;


	/* END ******************* CHORD SPECIFIC  CONSTANTS **************/
	
	/* *********** STATIC ATTRIBUTES FOR getMaximum() METHOD ********/
	/**
	 * Sets the number of bits used in last building of MAX.
	 */
	protected static int bitsKey;
	/**
	 * Inform the maximum value for this ChordId using the number of 
	 * bits <b>bitsKey</b>.
	 */
	protected static Id MAX;
	/* END *********** STATIC ATTRIBUTES FOR getMaximum() METHOD ********/
	

  /**
   * number of elements in the array
   */ 
  private int arrayElems = Properties.bitsKey / 32;

  /**
   * the id itself.
   */
  private int[] Id;

  
  /**
   * Generates an Id initialized to zero.
   * It is not required for the framework.
   */
  public ChordId() {
    Id = new int[arrayElems];

    for (int i = 0; i < arrayElems; i++) {
      Id[i] = 0;
    }
  }
  
  /**
   * This constructor is nonapplicable to the Chord overlay.
   * @param data int value
   * @throws InitializationException always that this constructor
   * is used.
   */
  public ChordId(int data) throws InitializationException {
  	throw new InitializationException("This constructor is nonapplicable.");
  }

  /**
   * This constructor is nonapplicable to the Chord overlay.
   * @param data double value
   * @throws InitializationException always that this constructor
   * is used.
   */
  public ChordId(double data) throws InitializationException {
  	throw new InitializationException("This constructor is nonapplicable.");
  }
  
  /**
   * Generates an Id from data in byte array format.
   * Before construct the target Id, the byte array is converted 
   * to int array.
   * @param data The array with the desired values of the new Id.
   */
  public ChordId (byte[] data) {
    Id = new int[arrayElems];
    
    translate(Utilities.toIntArray(data));
  }
  
  /**
   * Generates an Id from data in int array format.
   * @param data The array with the desired values of the new Id.
   */
  public ChordId (int[] data) {
    Id = new int[arrayElems];
    
    translate(data);
  }
  
  /**
   * This constructor generate a BigInteger with the number
   * inside of <b>material</B>, and then, translate it
   * to the correct Id.
   * @param material String representation of any desired Id.
   */
  public ChordId (String material) {
	parseBigInteger(new BigInteger(material));
  }
  /**
   * This constructor translate the BigInteger <b>value</b>
   * to the correct Id.
   * @param value BigInteger representation of the
   * any desired Id.
   */
  public ChordId(BigInteger value) {
	parseBigInteger(value);
  }
  
  /**
   * Capture the MSB (More Significative Bits) of the identification,
   * in fragments of MODULE bits (32 bits). The MSB remain the MSB 
   * at new identification and are discarded the LSB (Less Significative
   * Bits).
   * @param dataForId Array with any size, with fragments of representation
   * of MODULE bits.
   */
  private void translate(int[] dataForId) {
  	int dataLength = this.getLastIndex(dataForId)+1;
  	if (arrayElems == dataLength) {
  		for (int i =0; i< arrayElems; i++) {
  			Id[i] = dataForId[i];
  		}
  	} else if (arrayElems < dataLength) {
  		int length = dataLength-1;
  		int arrayLength = arrayElems-1;
  	  	for (int i = 0; (i < arrayElems); i++) {
  				Id[arrayLength-i] = dataForId[length-i];
  		}
  	} else { //arrayElems > dataLength
  		int i = 0;
  		//capture first elements from dataForId
  		for (i=0; i<dataLength; i++) {
  			Id[i] = dataForId[i];
  		}
  		//initialize the rest to zero
  		for (i++;i< arrayElems; i++) {
  			Id[i] = 0;
  		}
  	}
  }
  
  /**
   * Returns the last index with a value different to zero.
   * @param data Array to analyze.
   * @return The last index of array with a value different to zero.
   */
  private int getLastIndex(int[] data) {
  	int i =data.length-1;
  	while (i>0 && data[i]==0) i--;
  	return i;
  }


  
  /**
   * Translate the BigInteger <b>value</b> to an Id,
   * partitioning it to fragments of 32 bits. 
   */
  protected void parseBigInteger(BigInteger value) {
	Id = new int[arrayElems];
	int[] data = new int[MAX_ARRAY_SIZE];
	
  	BigInteger two = new BigInteger("2");
	BigInteger div = two.pow(MODULE);	
	
	//build entire identification with maximum precition
	for (int i = 0; i < MAX_ARRAY_SIZE; i++) {	  	
      BigInteger[] result = value.divideAndRemainder(div);
	  data[i] = result[1].intValue();	 
	  value = result[0];	  	  	   
	}
	
	//transform the entire identification to fragment 
	//for actual presition bits.
	translate(data);
  }

  /**
   * Returns an Id corresponding to this Id plus a given offset
   *
   * @param offset the offset to add
   * @return the new Id
   */
  public Id add (Id offset) {
    Id newId = new ChordId();
    long x;
    long y;
    long sum;
    int carry = 0;

    for (int i = 0; i < arrayElems; i++) {
      // Little-endian to big-endian conversion
      x = Id[i] & 0x0ffffffffL;
      y = ((ChordId) offset).Id[i] & 0x0ffffffffL;

      sum = x + y + carry;

      if (sum >= 0x100000000L) {
        carry = 1;
      } else {
        carry = 0;
      }

      ((ChordId) newId).Id[i] = (int) sum;
    }

    // Overflow: have to do modulus 2^bitLength
    if (carry == 1) {
    	newId = ((ChordId) newId).subtractWithCarry (getMaximum());
    }

    return newId;
  }

  public Id subtractWithCarry (Id offset) {
    Id newId = new ChordId();
    int carry = 1;
    long x;
    long y;
    long sub;

    for (int i = 0; i < arrayElems; i++) {
      x = Id[i] & 0x0ffffffffL;
      // Consider carry bit to be added to Id (it is in fact out of range!)
      if (i == arrayElems - 1) {
        x |= 0x100000000L;
      }
      y = ((ChordId) offset).Id[i] & 0x0ffffffffL;

      sub = x - y - carry;
      if (sub < 0) {
        carry = 1;
      }
      ((ChordId) newId).Id[i] = (int) sub;
    }

    return newId;
  }


  public Id subtract (Id offset) {
    Id newId = new ChordId();
    int carry = 0;
    long x;
    long y;
    long sub;

    for (int i = 0; i < arrayElems; i++) {
      x = Id[i] & 0x0ffffffffL;
      y = ((ChordId) offset).Id[i] & 0x0ffffffffL;

      sub = x - y - carry;
      if (sub < 0) {
        carry = 1;
      }
      ((ChordId) newId).Id[i] = (int) sub;
    }

    return newId;
  }

  public boolean equals (Object obj) {
    Id nid = (Id) obj;

    for (int i = 0; i < arrayElems; i++) {
      if (Id[i] != ((ChordId) nid).Id[i]) {
        return false;
      }
    }

    return true;
  }

  public void blit (byte target[]) {
    for (int j = 0; j < Properties.bitsKey / 8; j++) {
      int k = Id[j / 4] >> ((j % 4) * 8);
      target[j] = (byte) (k & 0xff);
    }
  }

  public byte[] copy() {
    byte target[] = new byte[Properties.bitsKey / 8];
    blit(target);
    return target;
  }

  /**
   * Comparison operator for Ids. The comparison that occurs is a numerical comparison.
   *
   * @param obj the Id to compare with.
   * @return negative if this < obj, 0 if they are equal and positive if this > obj.
   */
  public int compareTo (Object obj) {
    Id oth = (Id) obj;

    for (int i = arrayElems - 1; i >= 0; i--) {
      if (Id[i] != ((ChordId) oth).Id[i]) {
        long t = Id[i] & 0x0ffffffffL;
        long o = ((ChordId) oth).Id[i] & 0x0ffffffffL;
        if (t < o) {
          return -1;
        } else {
          return 1;
        }
      }
    }

    return 0;
  }

  /**
   * Returns the byte array representation of this Id
   *
   * @return The byte array representation of this id
   */
  public byte[] toByteArray() {
    return copy();
  }

  /**
   * Hash codes for Ids.
   *
   * @return a hash code.
   */
  public int hashCode() {
    int h = 0;

    /// Hash function is computed by XORing the bits of the Id.
    for (int i = 0; i < arrayElems; i++) {
      h ^= Id[i];
    }

    return h;
  }

  /**
   * Returns the complete represntation of this Id, in hex.
   *
   * @return The complete representation of this Id, in hexadecimal
   */
  public String toStringHexFull() {
    StringBuffer buffer = new StringBuffer();

    String tran[] = {"0", "1", "2", "3", "4", "5", "6", "7",
      "8", "9", "A", "B", "C", "D", "E", "F"};

    int n = Properties.bitsKey / 4;
    for (int i = n - 1; i >= 0; i--) {
      int d = getDigit(i, 4);

      buffer.append(tran[d]);
    }
    return buffer.toString();
  }

  public String toStringFull() {
    String hex = toStringHexFull();
    java.math.BigInteger dec;    // Decimal (base-10) equivalent of hexadecimal number.
    int i;                       // A position in hex, from 0 to hex.length()-1.
    dec = new java.math.BigInteger ("0");
    for ( i = 0; i < hex.length(); i++ ) {
      int digit = hexValue (hex.charAt(i));
      dec = ((dec.multiply (new java.math.BigInteger("16"))).add (new java.math.BigInteger (""+digit)));
    }
    return dec.toString();
  }

  private int hexValue (char ch) {
    // Returns the hexadecimal value of ch, or returns
    // -1 if ch is not one of the hexadecimal digits.
    switch (ch) {
      case '0':
        return 0;
      case '1':
        return 1;
      case '2':
        return 2;
      case '3':
        return 3;
      case '4':
        return 4;
      case '5':
        return 5;
      case '6':
        return 6;
      case '7':
        return 7;
      case '8':
        return 8;
      case '9':
        return 9;
      case 'a':
      case 'A':
        return 10;
      case 'b':
      case 'B':
        return 11;
      case 'c':
      case 'C':
        return 12;
      case 'd':
      case 'D':
        return 13;
      case 'e':
      case 'E':
        return 14;
      case 'f':
      case 'F':
        return 15;
      default:
        return -1;
    }
  }

  /**
   * Returns a string representation of the Id in base 16. The string is a byte string from most to
   * least significant.
   *
   * @return A String representation of this Id, abbreviated
   */
  public String toString() {
    return this.toStringFull();
  }

  /**
   * Gets the ith digit in base 2^b. i = 0 is the least significant digit.
   *
   * @param i which digit to get.
   * @param b which power of 2 is the base to get it in.
   * @return the ith digit in base 2^b.
   */
  public int getDigit(int i, int b) {
    int bitIndex = b * i + (Properties.bitsKey % b);
    int index = bitIndex / 32;
    int shift = bitIndex % 32;
    long val = Id[index];
    if (shift + b > 32) {
      val = (val & 0xffffffffL) | (((long) Id[index + 1]) << 32);
    }
    return ((int) (val >> shift)) & ((1 << b) - 1);
  }

  /** Shift operator. shift(-1,0) multiplies value of this by two, shift(1,0) divides by 2
   *
   * @param cnt the number of bits to shift, negative shifts left, positive shifts right
   * @param fill value of bit shifted in (0 if fill == 0, 1 otherwise)
   * @return this
   */
  public Id shift (int cnt, int fill) {
    return shift (cnt, fill, false);
  }


  /**
   * Shift operator. shift(-1,0) multiplies value of this by two, shift(1,0) divides by 2
   *
   * @param cnt the number of bits to shift, negative shifts left, positive shifts right
   * @param fill value of bit shifted in (0 if fill == 0, 1 otherwise)
   * @param roundUp if true, round up the results after right shifting
   * @return this
   */
  public Id shift (int cnt, int fill, boolean roundUp) {
    Id newId = new ChordId (this.Id);
    int carry = 0;
    int bit = 0;
    int lsb = 0;

    if (cnt > 0) {
      for (int j = 0; j < cnt; j++) {
        // shift right one bit
        carry = (fill == 0) ? 0 : 0x80000000;
        for (int i = arrayElems - 1; i >= 0; i--) {
          bit = ((ChordId) newId).Id[i] & 1;
          ((ChordId) newId).Id[i] = (((ChordId) newId).Id[i] >>> 1) | carry;
          carry = (bit == 0) ? 0 : 0x80000000;
        }

        if (j == 0) {
          lsb = bit;
        }
      }

      if (roundUp && lsb > 0) {
        newId = inc();
      }
    } else {
      for (int j = 0; j < -cnt; j++) {
        // shift left one bit
        carry = (fill == 0) ? 0 : 1;
        for (int i = 0; i < arrayElems; i++) {
          bit = ((ChordId) newId).Id[i] & 0x80000000;
          ((ChordId) newId).Id[i] = (((ChordId) newId).Id[i] << 1) | carry;
          carry = (bit == 0) ? 0 : 1;
        }
      }
    }
    return newId;
  }

  private Id inc() {
    Id newId = new ChordId();
    long x;
    long sum;
    int carry = 1;

    // add one
    for (int i = 0; i < arrayElems; i++) {
      x = Id[i] & 0x0ffffffffL;
      sum = x + carry;
      if (sum >= 0x100000000L) {
        carry = 1;
      } else {
        carry = 0;
      }
      ((ChordId) newId).Id[i] = (int) sum;
    }
    return newId;
  }

  public boolean between (Id ccw, Id cw) {
        if (ccw.equals(cw)) {
          return true;
        }
        else if (ccw.compareTo(cw) < 0) {
          return ((this.compareTo (ccw) > 0 ) && (this.compareTo (cw) < 0));
        }
        else {
          return ((this.compareTo (ccw) > 0) || (this.compareTo (cw) < 0));
        }
  }

  /**
   * Checks to see if the Id nid is clockwise or counterclockwise from this, on the ring. An Id is
   * clockwise if it is within the half circle clockwise from this on the ring. An Id is considered
   * counter-clockwise from itself.
   *
   * @param nid The Id we are comparing to
   * @return true if clockwise, false otherwise.
   */
  public boolean clockwise (Id nid) {
    boolean diffMSB = ((Id[arrayElems - 1] & 0x80000000) != (((ChordId) nid).Id[arrayElems - 1] & 0x80000000));
    int x;
    int y;
    int i;

    if ((x = (Id[arrayElems - 1] & 0x7fffffff)) != (y = (((ChordId) nid).Id[arrayElems - 1] & 0x7fffffff))) {
      return ((y > x) ^ diffMSB);
    } else {
      for (i = arrayElems - 2; i >= 0; i--) {
        if (Id[i] != ((ChordId) nid).Id[i]) {
          break;
        }
      }

      if (i < 0) {
        return diffMSB;
      } else {
        long xl;
        long yl;

        xl = Id[i] & 0xffffffffL;
        yl = ((ChordId) nid).Id[i] & 0xffffffffL;

        return ((yl > xl) ^ diffMSB);
      }
    }
  }  
  
	/**
	 * Return the int[] with the value of this Id.
	 * @see planet.commonapi.Id#getValue()
	 * @return int[] with the value of this Id.
	 */
	public Object getValue() {
		return Id;
	}
	
	/**
	 * Sets the new value with the int[] that appears at <b>newValue</b>.
	 * @see planet.commonapi.Id#setValue(java.lang.Object)
	 * @param newValue New value as int[] representation.
	 */
	public void setValue(Object newValue) {
		translate((int[])newValue);
	}
	
	
	
	/**
	 * This method returns the maximum value for a ChordId that is 
     * possible to build, according the number of bits for key
     * actually in use. To return this maximum value use the
     * design pattern Singleton, reviewing the number of bits in use. 
     * @return The maximum value sets to an Id, always according
     * the number of bits actually in use.
	 * @see planet.commonapi.Id#getMaximum()
	 */
	public static Id getMaximum() {
		if (MAX == null || Properties.bitsKey != bitsKey) {
			bitsKey = Properties.bitsKey; //update values
			MAX = new ChordId(MAX_CHORD); 
		}
		return MAX;
	}
}