package planet.chord;
import java.math.BigInteger;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import planet.chord.message.BroadcastMessage;
import planet.chord.message.IdMessage;
import planet.chord.message.NodeMessage;
import planet.chord.message.SuccListMessage;
import planet.commonapi.EndPoint;
import planet.commonapi.Id;
import planet.commonapi.Message;
import planet.commonapi.NodeHandle;
import planet.commonapi.RouteMessage;
import planet.commonapi.exception.InitializationException;
import planet.generic.commonapi.factory.GenericFactory;
import planet.simulate.Globals;
import planet.simulate.Logger;
import planet.simulate.MessageListener;
import planet.simulate.MessagePool;
import planet.simulate.Results;
import planet.util.Properties;
import planet.util.Queue;
import planet.util.QueueFull;
import planet.util.timer.TimerTaskImpl;
/**
 * 
 * A Chord node is single entity in the chord network. It extends of the class
 * Node and specializes following the lookup Chord protocol. Moreover, the
 * stabilization implementation, producing the periodic stabilition events.
 * 
 * @author <a href="mailto:Ruben.Mondejar@estudiants.urv.es">Ruben Mondejar
 *         </a>
 * @author <a href="mailto:cpairot@etse.urv.es">Carles Pairot </a>
 * @author <a href="mailto:jordi.pujol@estudiants.urv.es">Jordi Pujol </a>
 *  
 */
public class ChordNode extends planet.generic.commonapi.NodeImpl {
	
	/* ******************  CONSTANTS FOR MODE OF ROUTEMESSAGE *******/
	
	/**
	 * Mode value: Defines the start of communication that requires 
	 * response (REPLY). 
	 */
	public final static int REQUEST		= 0;
	/**
	 * Mode value: Defines the response of a communication.
	 */
	public final static int REPLY		= 1;
	/**
	 * Mode value: Defines a message's mode that requires only
	 * communication on one way.
	 */
	public final static int REFRESH  	= 2;		
	
	/* END ******************  CONSTANTS FOR MODE OF ROUTEMESSAGE *******/
	
	/* ******************  CONSTANTS FOR TYPE OF ROUTEMESSAGE *******/
	/**
	 * Type value: First message that is send by any node to get its
	 * inmediate successor.
	 */
	public final static int FIND_SUCC 	= 3;
	/**
	 * Type value: Start a set of messages that permits to find the
	 * inmediate successor of any new incoming node to the network.
	 */
	public final static int FIND_PRE   	= 4;
	/**
	 * Type value: Informs to the node that receives this message its
	 * new inmediate successor.
	 */
	public final static int SET_SUCC   	= 5;
	/**
	 * Type value: Informs to the node that receive this message its
	 * new inmediate predecessor. This type forces to node to change its
	 * predecessor.
	 */
	public final static int SET_PRE    	= 6;
	/**
	 * Type value: Start a set of messages that permits to find the
	 * inmediate node in charge of any key.
	 */
	public final static int GET_PRE    	= 7;
	/**
	 * Type value: Informs to the node that receive this message its
	 * inmediate predecessor. This type permits to the node to evaluate
	 * if this notification informs really of the predecessor of the node.
	 */
	public final static int NOTIFY     	= 8;
	/**
	 * Type value: Start a set of messages that permits the harvesting of
	 * successors of the node that send originally this message.
	 */
	public final static int SUCC_LIST  	= 9;
	/**
	 * Type value: The message with this type identifies a broadcast message.
	 */
	public final static int BROADCAST  	= 10;
	/**
	 * Type value: It identifies that this message contains an application 
	 * level Message. 
	 */
	public final static int DATA 		= 11;	
	
	/* END ******************  CONSTANTS FOR TYPE OF ROUTEMESSAGE *******/
	
	/* ******************  CONSTANTS FOR TYPE/MODE OF ROUTEMESSAGE *******/
	/**
	 * This String contains a string representation of each of types
	 * and modes values of the RouteMessage.
	 */
	public final static String[] TYPES = {"REQUEST","REPLY","REFRESH",
			"FIND_SUCC","FIND_PRE","SET_SUCC","SET_PRE","GET_PRE","NOTIFY",
			"SUCC_LIST","BROADCAST","DATA"}; 
	
	/* END ******************  CONSTANTS FOR TYPE/MODE OF ROUTEMESSAGE *******/
	
	/**
	 * Return a RouteMessage with the specified values. If there are RouteMessages free,
	 * just build a new one. For generate new instances of RouteMessage will be used
	 * the implementation class that appears in properties file.
	 * @param key Identification of communication.
	 * @param from Source node.
	 * @param to Destination node.
	 * @param nextHop NextHop node.
	 * @return A RouteMessage with the specified values.
	 */
	protected static RouteMessage getDataMessage(String appId,
			NodeHandle from, NodeHandle to, NodeHandle nextHop, Message msg) 
				throws InitializationException {
	  	//getMessage(String key, 
	  	//NodeHandle from, NodeHandle to, NodeHandle nextHop, Message msg, 
	  	//int type, int mode, String appId)
	  	RouteMessage toReturn = MessagePool.getMessage("",from,to,nextHop,null,DATA,REQUEST,appId);
	  	toReturn.setMessage(msg);
	  	return toReturn;
	}
	
	/**
	 * Return a RouteMessage with the specified values. If there are RouteMessages free,
	 * just build a new one. For generate new instances of RouteMessage will be used
	 * the implementation class that appears in properties file.
	 * @param key Identification of communication.
	 * @param from Source node.
	 * @param to Destination node.
	 * @param nextHop NextHop node.
	 * @return A RouteMessage with the specified values.
	 */
	protected static RouteMessage getBroadcastMessage(String appId,
			NodeHandle from, NodeHandle to, NodeHandle nextHop, Message msg) 
				throws InitializationException {
		//getMessage(String key, 
		//NodeHandle from, NodeHandle to, NodeHandle nextHop, Message msg, 
		//int type, int mode, String appId)
	  	RouteMessage toReturn = MessagePool.getMessage("",from,to,nextHop,null,BROADCAST,REFRESH,appId);
	  	toReturn.setMessage(msg);
	  	return toReturn;
	}
	
	
	
	protected Id MAX;
	protected int null_pointers = 0;
	/**
	 * Detect the number of finger changes only on one simulation step.
	 */
	protected int finger_changes = 0;
	/**
	 * Number of simulation steps without changes on finger table.
	 */
	protected int stabRate = 0;
	/**
	 * Number of simulation steps without changes on finger table, that
	 * represents a real stabilization of this node.
	 */
	protected int realStabilizationRate = 
		((ChordProperties)Properties.overlayProps).stabilizeSteps *
				planet.util.Properties.bitsKey;
	protected NodeHandle[] finger;
	protected Id[] start;
	protected NodeHandle predecessor;
	protected Vector succ_list;
	protected boolean hasFailed; //for failed case
	protected boolean hasLeaved; //for leaved case
	protected NodeHandle[] temp;
	/**
	 * Temporal use in Common API methods.
	 */
	protected Vector auxCAPI;
	public Id deux;
	/**
	 * Local EndPoints.
	 */
	//protected Hashtable endpoints;
	/**
	 * Timer to schedule differents tasks.
	 */
	//protected Timer timer;

	/**
	 * Constructor, create a new BadChordNode instance with this node Id
	 * 
	 * @param id
	 *            Id to set to the new node
	 */
	public ChordNode(Id id) throws InitializationException {
		super(id);
		deux = GenericFactory.buildId(ChordId.TWO_CHORD);
		finger = new NodeHandle[Properties.bitsKey];
		start = new Id[Properties.bitsKey];
		MAX = GenericFactory.buildId(ChordId.MAX_CHORD);
		temp = new NodeHandle[3];
		endpoints = new Hashtable();
		for (int i = 0; i < Properties.bitsKey; i++) {
			start[i] = (id.add(deux.shift(-(i - 1), 0)));
		}
		succ_list = new Vector();
		hasLeaved = false;
		hasFailed = false;
		//add stabilize timer
		setTimer(new StabilizeTask(),  ((ChordProperties)Properties.overlayProps).stabilizeSteps,
				((ChordProperties)Properties.overlayProps).stabilizeSteps);
		// add fix finger timer
		setTimer(new FixFingerTask(),  ((ChordProperties)Properties.overlayProps).fixFingerSteps,
				((ChordProperties)Properties.overlayProps).fixFingerSteps);
	}
	
	
	
	/**
	 * Sets the predecessor node
	 * 
	 * @param handle
	 *            NodeHandle of the predecessor
	 */
	public void setPred(NodeHandle handle) {
		predecessor = handle;
	}
	/**
	 * Sets the successor node
	 * 
	 * @param handle
	 *            NodeHandle of the successor
	 */
	public void setSucc(NodeHandle handle) {
		finger[0] = handle;
	}
	/**
	 * Returns the NodeHandle of the present identification of the precursor of the
	 * node
	 * 
	 * @return predeccesor NodeHandle
	 */
	public NodeHandle getPred() {
		return predecessor;
	}
	/**
	 * Returns the NodeHandle of the present identification of the successor of the
	 * node
	 * 
	 * @return predeccesor NodeHandle
	 */
	public NodeHandle getSucc() {
		return finger[0];
	}
	/**
	 * Returns the successor list of the present node
	 * 
	 * @param max
	 *            number of the elements to return
	 * @return successor list Vector
	 */
	public Vector getSuccList(int max) {
		Iterator it = succ_list.iterator();
		Vector sl = new Vector();
		int i = 0;
		while (it.hasNext() && i < max)
			sl.add(it.next());
		return sl;
	}
	
	/**
	 * Clears the counter of the finger table changes
	 */
	protected void clearFingerChanges() {
		finger_changes = 0;
	}
	/**
	 * Returns the count of the finger table changes
	 * 
	 * @return finger_changes int counter
	 */
	protected int getFingerChanges() {
		return finger_changes;
	}
	/**
	 * Sets a especific finger of the finger table
	 * 
	 * @param position
	 *            int position of the finger
	 * @param handle
	 *            new NodeHandle to change
	 */
	protected void setFinger(int pos, NodeHandle handle) {
		if (finger[pos] == null || !finger[pos].equals(handle)) {
			finger_changes++;
			finger[pos] = handle;
			//Results.incStabRate();
		}
	}
	
	public Hashtable getInfo() {
		Hashtable info = new Hashtable();
		info.put("predecessor", predecessor);
		info.put("finger", finger);
		info.put("start", start);
		info.put("succ_list", succ_list);
		return info;
	}
	/**
	 * Shows for System.out all information of the node, including finger
	 * table.
	 */
	public void printNode() {
		System.out.println("######\nNode:        " + getId());
		System.out.println("Finger table:");
		for (int i = 0; i < finger.length; i++) {
			if (finger[i] != null)
				System.out.println(start[i] + " : " + finger[i].getId());
			else
				System.out.println(" finger [" + i + "] == null ");
			System.out.println("-----------");
		}
		System.out.println("Predecessor: " + predecessor.getId());
		System.out.println("Succ-List: " + succ_list);
		System.out.println("###### END OF NODE\n");
	}
	/**
	 * Shows for System.out only the owner Id, its predecessor and successor.
	 */
	public void prettyPrintNode() {
		System.out.println("######\nNode:        " + getId());
		System.out.println("Successor:   " + finger[0].getId());
		System.out.println("Predecessor: " + predecessor.getId());
		System.out.println("###### END OF NODE\n");
	}
	/**
	 * Finds the successor of a node. If the result if null because the
	 * information its not contain in the finger table, it generate a request
	 * message and the response message content the real result
	 * 
	 * @param handle
	 *             NodeHandle of the one node
	 * @return successor Id of the node successor
	 */
	protected NodeHandle find_successor(NodeHandle handle) {
		if (predecessor != null && handle.getId().betweenE(predecessor.getId(), this.id)) {
			return this.nodeHandle;
		} else {
			return find_predecessor(handle);
		}
	}
	/**
	 * Finds the predecessor of a node. If the result if null because the
	 * information its not contain in the finger table, it generate a request
	 * message and the response message content the real result
	 * 
	 * @param handle
	 *            NodeHandle of the one node
	 * @return successor NodeHandle of the node predecessor
	 */
	protected NodeHandle find_predecessor(NodeHandle handle) {
		if (handle.getId().equals(this.id)) {
			return this.nodeHandle; //return predecessor.successor();
		} else if (finger[0] != null && handle.getId().betweenE(this.id, finger[0].getId())) {
			return finger[0];
		} else {
			return null;
		}
	}
	
	/**
	 * Finds the closest preceding finger of a node
	 * 
	 * @param id
	 *            Id of the one node
	 * @return successor NodeHandle of the node successor
	 */
	protected NodeHandle closest_preceding_finger(Id id) {
		boolean found = false;
		temp[2] = this.nodeHandle; //n_id
		for (int i = Properties.bitsKey - 1; (i > -1 && !found); i--) {
			if (finger[i] != null && finger[i].getId().between(this.id, id)) {
				temp[2] = finger[i];
				found = true;
			}
		}
		return temp[2];
	}
	/**
	 * Finds the successor of a node. If the result if null because the
	 * information its not contain in the finger table, it generate a request
	 * message and the the content response message has been store in the
	 * finger table
	 * 
	 * @param pos
	 *            position of NodeHandle in the start table
	 * @return successor NodeHandle of the node successor
	 */
	protected NodeHandle find_successor(int pos) {
		if (predecessor != null && start[pos].betweenE(predecessor.getId(), this.id)) {
			return this.nodeHandle;
		} else {
			return find_predecessor(pos);
		}
	}
	/**
	 * Finds the predecessor of a node. If the result if null because the
	 * information its not contain in the finger table, it generate a request
	 * message and the the content response message has been store in the
	 * finger table
	 * 
	 * @param pos
	 *            position of NodeHandle in the start table
	 * @return successor NodeHandle of the node predecessor
	 */
	protected NodeHandle find_predecessor(int pos) {
		if (start[pos].equals(this.id)) {
			return this.nodeHandle; //return predecessor.successor();
		} else if (finger[0] != null && start[pos].betweenE(this.id, finger[0].getId())) {
			return finger[0];
		} else {
			RouteMessage aMsg = null;
			try {
				temp[1] = closest_preceding_finger(start[pos]); //n1_id
				String key = GenericFactory.generateKey(); //keyGen.generateKey();
				aMsg = MessagePool.getMessage(key, this.nodeHandle, temp[1],
						FIND_PRE, REQUEST);
				aMsg.setMessage(new IdMessage(start[pos]));
				Hashtable info = new Hashtable();
				info.put("Position", new Integer(pos));
				addMessageListener(key, new FindPredListener(this, info));
				send(aMsg);
			} catch (QueueFull e) {
				Logger.log("Outgoing Queue of Node " + this.id + " is Full",
						Logger.ERROR_LOG);
				MessagePool.freeMessage(aMsg);
			} catch (InitializationException e) {
				Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n"
						+ e.getMessage(), Logger.ERROR_LOG);
			}
		}
		//	return old finger result till received the correct value
		return finger[pos];
	}
	/**
	 * The node joins in the network
	 * 
	 * @param bootstrap
	 *            Id of arbitrary node in the network
	 */
	public void join(Id n1Id) {
		//System.out.println("\nNode ["+id+"] join to ["+n1Id+"]; step ["+ControlTimer.currentStep+"]\n");
		if (n1Id.equals(id)) {
			for (int i = 0; i < Properties.bitsKey; i++) {
				finger[i] = this.nodeHandle;
			}
			predecessor = this.nodeHandle;
		} else {
			RouteMessage aMsg = null;
			try {
				predecessor = null;
				String key = GenericFactory.generateKey(); //keyGen.generateKey();
				aMsg = MessagePool.getMessage(key, nodeHandle, 
						GenericFactory.buildNodeHandle(n1Id,true), FIND_SUCC,
						REQUEST);
				addMessageListener(key, new FindSuccListener(this));
				send(aMsg);
			} catch (QueueFull e) {
				Logger.log("Outgoing Queue of Node " + this.id + " is Full",
						Logger.ERROR_LOG);
				MessagePool.freeMessage(aMsg);
			} catch (InitializationException e) {
				Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n"
						+ e.getMessage(), Logger.ERROR_LOG);
			}
		}
	}
	
	/**
	 * The node leaves the network
	 */
	public void leave() {
		hasLeaved = true;
		RouteMessage aMsg = null;
		try {
			aMsg = MessagePool.getMessage(null, this.nodeHandle, predecessor,
					SET_SUCC, REFRESH);
			aMsg.setMessage(new NodeMessage(finger[0]));
			send(aMsg);
		} catch (QueueFull e) {
			Logger.log("Outgoing Queue of Node " + this.id + " is Full",
					Logger.ERROR_LOG);
			MessagePool.freeMessage(aMsg);
		} catch (InitializationException e) {
			Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n"
					+ e.getMessage(), Logger.ERROR_LOG);
		}
		RouteMessage bMsg = null;
		try {
			bMsg = MessagePool.getMessage(null, this.nodeHandle, finger[0],
					SET_PRE, REFRESH);
			bMsg.setMessage(new NodeMessage(predecessor));
			send(bMsg);
		} catch (QueueFull e) {
			Logger.log("Outgoing Queue of Node " + this.id + " is Full",
					Logger.ERROR_LOG);
			MessagePool.freeMessage(bMsg);
		} catch (InitializationException e) {
			Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n"
					+ e.getMessage(), Logger.ERROR_LOG);
		}
		Logger.log("LEAVING NODE " + id, Logger.EVENT_LOG);
	}
	/**
	 * Verify node immediate successor and tell the successor about this,
	 * moreover generate and update the successor list.
	 */
	protected void stabilize() {
		String key = GenericFactory.generateKey(); //keyGen.generateKey();
		RouteMessage aMsg = null;
		try {
			aMsg = MessagePool.getMessage(key, nodeHandle, finger[0], GET_PRE,
					REQUEST);
			addMessageListener(key, new GetPreListener(this));
			send(aMsg);
		} catch (QueueFull e) {
			Logger.log("Outgoing Queue of Node " + this.id + " is Full",
					Logger.ERROR_LOG);
			MessagePool.freeMessage(aMsg);
		} catch (InitializationException e) {
			Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n"
					+ e.getMessage(), Logger.ERROR_LOG);
		}
		//if the links are been correct then sends the request to successor
		if (finger[0] != null && !finger[0].equals(this.nodeHandle)
				&& predecessor != null && !finger[0].equals(predecessor)) {
			RouteMessage bMsg = null;
			try {
				bMsg = MessagePool.getMessage(null, nodeHandle, finger[0],
					SUCC_LIST, REQUEST);
				bMsg.setMessage(new SuccListMessage());
				send(bMsg);
			} catch (QueueFull e) {
				Logger.log("Outgoing Queue of Node " + this.id + " is Full",
						Logger.ERROR_LOG);
				MessagePool.freeMessage(bMsg);
			} catch (InitializationException e) {
				Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n"
						+ e.getMessage(), Logger.ERROR_LOG);
			}
		}
	}
	/**
	 * Verify if the indicated node is the true predecessor
	 * 
	 * @param id
	 *            node to check
	 */
	protected void notify(NodeHandle nh) {
		if (predecessor == null || nh.getId().between(predecessor.getId(), this.id)) {
			predecessor = nh;
		}
	}
	/**
	 * Finds the predecessor of a node. If the result if null because the
	 * information its not contain in the finger table, it generate a request
	 * message and the the content response message has been store in the
	 * finger table
	 * 
	 * @param position
	 *            position of Id in the start table
	 * @return successor Id of the node predecessor
	 */
	//periocally refresh finger table entries
	protected void fix_fingers() {
		if (null_pointers == 0) {
			null_pointers++;
		}
		setFinger(null_pointers, find_successor(null_pointers));
		null_pointers = (null_pointers + 1) % Properties.bitsKey;
	}
	/**
	 * Given a key, it generates the message and listener to obtain the owner
	 * of this one.
	 * 
	 * @param key
	 *            Id of the key
	 */
	public void lookup(Id id) {
		try {
			temp[0] = find_successor(GenericFactory.buildNodeHandle(id,true)); //succ
		} catch (InitializationException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		if (temp[0] != null) {
			try {
				RouteMessage aMsg = MessagePool.getMessage(null, this.nodeHandle,
						temp[0], FIND_PRE, REPLY);
				//Results.updateHopsMsg(this.id,keyGen.generateKey());
				Results.updateHopsMsg(this.id, GenericFactory.generateKey());
				aMsg.setMessage(new NodeMessage(temp[0]));
				(new LookupListener(this, id)).onMessage(aMsg);
			} catch (InitializationException e) {
				Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n"
						+ e.getMessage(), Logger.ERROR_LOG);
			}
		} else {
			String key = GenericFactory.generateKey(); //keyGen.generateKey();
			NodeHandle toFind = closest_preceding_finger(id);
			RouteMessage aMsg = null;
			try {
				aMsg = MessagePool.getMessage(key, this.nodeHandle, toFind,
						FIND_PRE, REQUEST);
				Logger.log("NODE " + this.id + " LOOKUP --> " + id
						+ " message key : " + key, Logger.MSG_LOG);
				aMsg.setMessage(new IdMessage(id));
				addMessageListener(key, new LookupListener(this, id));
				send(aMsg);
			} catch (QueueFull e) {
				Logger.log("Outgoing Queue of Node " + this.id + " is Full",
						Logger.ERROR_LOG);
				MessagePool.freeMessage(aMsg);
			} catch (InitializationException e) {
				Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n"
						+ e.getMessage(), Logger.ERROR_LOG);
			}
		}
	}
	
	/**
	 * Send a message to destination node directly.
	 * 
	 * @param idMessage
	 *            Id of message
	 * @param message
	 *            Message to deliver from application.
	 * @param id
	 *            Id of destination node
	 */
	protected void sendData(RouteMessage message, NodeHandle hint) {
		message.setKey(GenericFactory.generateKey());
		message.setType(DATA);
		message.setMode(REFRESH);
		message.setNextHopHandle(hint);
		try {
			send(message);
			Results.incTraffic();
		} catch (QueueFull e) {
			Logger.log("Outgoing Queue of Node " + id + " is Full",
					Logger.ERROR_LOG);
			MessagePool.freeMessage(message);
		}
	}
	/**
	 * Send a message to unknown destination node via routing.
	 * 
	 * @param idMessage
	 *            Message's Id.
	 * @param message
	 *            Message to deliver from application.
	 */
	protected void routingData(RouteMessage message, NodeHandle nextHop) {
		//enrutar missatge via key del missatge (el idMessage)
		if (predecessor != null && nextHop.getId().betweenE(predecessor.getId(), this.id)) {
			//deliver
			EndPoint endpoint = (EndPoint) endpoints.get(message
					.getApplicationId());
			endpoint.scheduleMessage(message, 0);
		} else if (finger[0] != null && nextHop.getId().betweenE(this.id, finger[0].getId())) {
			sendData(message, finger[0]);
		} else {
			NodeHandle toFind = closest_preceding_finger(nextHop.getId());
			message.setKey(GenericFactory.generateKey());
			message.setType(DATA);
			message.setMode(REQUEST);
			message.setNextHopHandle(toFind); //sets the destination Id
			//the Message has been set for EndPoint
			try {
				send(message);
				Results.incTraffic();
				Results.updateHopsMsg(message.getSource().getId(), message.getKey());
			} catch (QueueFull e) {
				Logger.log("Outgoing Queue of Node " + id + " is Full",
						Logger.ERROR_LOG);
				MessagePool.freeMessage(message);
			}
		}
	}
	
	/**
	 * Sends a RouteMessage to the ring. If nextHop is null, it shows that the
	 * destination node is unknown and the message will be routed. If nextHop
	 * is not null, it's the destination NodeHandle and it must be reached
	 * directly.
	 * 
	 * @param data
	 *            The RouteMessage to be routed to the ring.
	 *  
	 */
	public void routeData(String appId,
			NodeHandle to, NodeHandle nextHop, Message msg) {
		RouteMessage toSend;
		try {
			toSend = getDataMessage(appId,nodeHandle,to,nextHop,msg);
			if (nextHop != null) {
				this.sendData(toSend, nextHop);
			} else {
				this.routingData(toSend, to);
			}
		} catch (InitializationException e) {
			Logger.log("["+id+"] cannot build a RouteMessage for this data ["+
					msg+"]",Logger.ERROR_LOG);
		}
	}

	/**
	 * Initiating a Broadcast Message
	 * 
	 * @param info
	 *            Hashtable with the information to transmit
	 */
	public void broadcast(String appId,
			NodeHandle to, NodeHandle nextHop, Message msg) {
		NodeHandle limit;
		NodeHandle r;
		for (int i = 0; i < Properties.bitsKey - 1; i++) {
			//Skip a redundant finger
			if (!finger[i].equals(finger[i + 1])) {
				r = finger[i];
				limit = finger[i + 1];
				RouteMessage aMsg = null;
				try {
					//String appId,Id from, Id to, Id nextHop
					aMsg = getBroadcastMessage(appId,
							this.nodeHandle, r, r,new BroadcastMessage(msg, limit));
					send(aMsg);
				} catch (QueueFull e) {
					Logger.log(
							"Outgoing Queue of Node " + this.id + " is Full",
							Logger.ERROR_LOG);
					MessagePool.freeMessage(aMsg);
				} catch (InitializationException e) {
					Logger.log("Cannot build a new instance of RouteMessage for BroadcastMessage",
									Logger.ERROR_LOG);
				}
			}
		}
		//Process the last finger
		RouteMessage aMsg = null;
		try {
			//String appId,Id from, Id to, Id nextHop
			aMsg = getBroadcastMessage(appId, this.nodeHandle,
					finger[Properties.bitsKey - 1],
					finger[Properties.bitsKey - 1],
					new BroadcastMessage(msg, nodeHandle));
			//I'm the limit of the last finger
			send(aMsg);
		} catch (QueueFull e) {
			Logger.log("Outgoing Queue of Node " + this.id + " is Full",
					Logger.ERROR_LOG);
			MessagePool.freeMessage(aMsg);
		} catch (InitializationException e) {
			Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n"
					+ e.getMessage(), Logger.ERROR_LOG);
		}
	}
	/**
	 * Treats the messages and according to the case, executes the generic
	 * listeners or listeners specialized, forward the messages or send
	 * responses messages
	 * 
	 * @param msg
	 *            IMessage to treat
	 *  
	 */
	public void dispatcher(RouteMessage msg) {
		//Response message but not successor list type (key == null)
		if (msg.getMode() == REPLY && msg.getKey() != null) {
			String key = msg.getKey();
			try {
				((MessageListener) listeners.get(key)).onMessage(msg);
			} catch (NullPointerException e) {
				Logger.log("I'm [" + id + "]; not exist listener for key ["
						+ msg.getKey() + "]", Logger.ERROR_LOG);
			}
			removeMessageListener(key);
		} else if (msg.getMode() == Globals.ERROR) {
			if (msg.getKey() != null) {
				String key_fp = msg.getKey();
				Logger.log("Node " + this.id + " destroy message key " + key_fp
						+ " type " + Globals.typeToString(msg.getType()) + " content "
						+ msg.getMessage(), Logger.MSG_LOG);
				MessageListener lst = (MessageListener) listeners.get(key_fp);
				if (lst != null) {
					removeMessageListener(key_fp);
				}
			}
			//Successor lost
			if (finger[0] != null && msg.getSource().equals(finger[0])
					&& succ_list.size() > 0) {
				//if not exists, succ_list is unchanged
				succ_list.remove(finger[0]);
				if (succ_list.size() > 0) {
					finger[0] = (NodeHandle) succ_list.firstElement();
					//send notify
					msg.setSource(this.nodeHandle);
					msg.setDestination(finger[0]);
					msg.setNextHopHandle(finger[0]);
					msg.setType(SET_PRE);
					msg.setMode(REFRESH);
					msg.setMessage(new NodeMessage(nodeHandle));
					try {
						send(msg);
					} catch (QueueFull e) {
						Logger.log("Outgoing Queue of Node " + this.id
								+ " is Full", Logger.ERROR_LOG);
						MessagePool.freeMessage(msg);
					}
				}
			} else
				//if source not exists, succ_list is unchanged
				succ_list.remove(msg.getSource());
		} else {
			switch (msg.getType()) {
				//DATA
				case DATA :
					String id_ap = msg.getApplicationId();
					EndPoint endpoint = (EndPoint) endpoints.get(id_ap);

					if (msg.getMode() == REQUEST) {
						boolean forward = endpoint.forward(msg);

						if (forward) {
							if (finger[0] != null
									&& msg.getDestination().getId().betweenE(this.id,
											finger[0].getId())) {
								//found successor --> send data
								msg.setNextHopHandle(finger[0]);
								msg.setType(DATA);
								msg.setMode(REFRESH);
								try {
									send(msg);
									Results.updateHopsMsg(msg.getSource().getId(),
											msg.getKey());
								} catch (QueueFull e) {
									Logger.log("Outgoing Queue of Node "
											+ this.id + " is Full",
											Logger.ERROR_LOG);
									MessagePool.freeMessage(msg);
								}
							} else {
								//next node [ROUTING]
								msg.setNextHopHandle(closest_preceding_finger(
										msg.getDestination().getId()));
								msg.setType(DATA);
								msg.setMode(REQUEST);
								try {
									send(msg);
									Results.updateHopsMsg(msg.getSource().getId(),
											msg.getKey());
								} catch (QueueFull e) {
									Logger.log("Outgoing Queue of Node "
											+ this.id + " is Full",
											Logger.ERROR_LOG);
									MessagePool.freeMessage(msg);
								}
							}
						} else {
							Results.decTraffic();
						}
					} else if (msg.getMode() == REFRESH) {
						//	    	Deliver message to End Point
						boolean forward = endpoint.forward(msg);
						if (forward) {
							id_ap = msg.getApplicationId();
							endpoint = (EndPoint) endpoints.get(id_ap);
							endpoint.scheduleMessage(msg, 0);
							Results.decTraffic();
						} else {
							Results.decTraffic();
						}
					} else
						Results.decTraffic();
					break;
				//CONTROL
				//REFRESH
				case SET_SUCC :
					NodeHandle succ = ((NodeMessage) msg.getMessage()).getNode();
					setSucc(succ);
					succ_list.remove(succ);
					break;
				case SET_PRE :
					setPred(((NodeMessage) msg.getMessage()).getNode());
					break;
				case NOTIFY :
					notify(msg.getSource());
					break;
				case BROADCAST :
					NodeHandle r, new_limit;
					BroadcastMessage bm = (BroadcastMessage) msg.getMessage();
					NodeHandle limit = bm.getLimit();
					Id limitId = limit.getId();
					planet.commonapi.Message info = bm.getInfo();
					
					for (int i = 0; i < Properties.bitsKey - 1; i++) {
						//Skip a redundant finger
						if (!finger[i].equals(finger[i + 1])) {
							//Forward while within "Limit"
							if (finger[i].getId().between(this.id, limitId)) {
								r = finger[i];
								//New Limit must not exceed Limit
								if (finger[i + 1].getId().between(this.id, limitId)) {
									new_limit = finger[i + 1];
								} else {
									new_limit = limit;
								}
								//no reuse of RouteMessage msg ==> send one
								// message to different nodes ==> requires
								// different messages
								planet.commonapi.RouteMessage aMsg = null;
								try {
									// String appId,Id from, Id to, Id nextHop
									aMsg = getBroadcastMessage(msg
											.getApplicationId(), this.nodeHandle, r, r,
											new BroadcastMessage(info,new_limit));
									send(aMsg);
								} catch (QueueFull e) {
									Logger.log("Outgoing Queue of Node " + this.id 
											+ " is Full", Logger.ERROR_LOG);
									MessagePool.freeMessage(aMsg);
								} catch (InitializationException e) {
									Logger.log(
											"ERROR: Cannot get a RouteMessage of MessagePool\n"
											+ e.getMessage(), Logger.ERROR_LOG);
								}
							}
						}
					}
					Logger.log("Broadcast : Node " + this.id + " info : "
							+ info, Logger.EVENT_LOG);
					msg.setMessage(info);
					((EndPoint) endpoints.get(msg.getApplicationId())).scheduleMessage(msg, 0);
					break;
				//REQUEST
				case FIND_SUCC :
					//source --> join();
					NodeHandle fSucc = find_successor(msg.getSource());
					if (fSucc != null) {
						msg.setMessage(new NodeMessage(fSucc));
						msg.setDestination(msg.getSource());
						msg.setNextHopHandle(msg.getSource());
						msg.setSource(nodeHandle);
						msg.setMode(REPLY);
						try {
							send(msg);
						} catch (QueueFull e) {
							Logger.log("Outgoing Queue of Node " + this.id
									+ " is Full", Logger.ERROR_LOG);
							MessagePool.freeMessage(msg);
						}
					} else {
						String key_fp = GenericFactory.generateKey(); //keyGen.generateKey();
						//order of invokes is important!!
						msg.setMessage(new IdMessage(msg.getSource().getId()));
						msg.setDestination(closest_preceding_finger(msg.getSource().getId()));
						msg.setNextHopHandle(msg.getDestination());
						msg.setSource(this.nodeHandle);
						String oldKey = msg.getKey();
						msg.setKey(key_fp);
						msg.setType(FIND_PRE);
						msg.setMode(REQUEST);
						try {
							Hashtable hashInfo = new Hashtable();
							hashInfo.put("Key", oldKey);
							addMessageListener(key_fp, new FindPredListener(this, hashInfo));
							send(msg);
						} catch (QueueFull e) {
							Logger.log("Outgoing Queue of Node " + this.id
									+ " is Full", Logger.ERROR_LOG);
							MessagePool.freeMessage(msg);
						}
					}
					break;
				case FIND_PRE :
					if (finger[0] != null
							&& ((IdMessage) msg.getMessage()).getNode().betweenE(
									this.id, finger[0].getId())) {
						//return successor
						Id msgId = ((IdMessage) msg.getMessage()).getNode();
						NodeHandle originalSource = msg.getSource();
						msg.setMessage(new NodeMessage(getSucc()));
						try {
							msg.setSource(GenericFactory.buildNodeHandle(msgId,true));
						} catch (InitializationException e1) {
							// TODO Auto-generated catch block
							e1.printStackTrace();
						}
						msg.setDestination(originalSource);
						msg.setNextHopHandle(originalSource);
						msg.setType(FIND_PRE);
						msg.setMode(REPLY);
						try {
							send(msg);
						} catch (QueueFull e) {
							Logger.log("Outgoing Queue of Node " + this.id
									+ " is Full", Logger.ERROR_LOG);
							MessagePool.freeMessage(msg);
						}
					} else {
						//next node
						Id aux = ((IdMessage) msg.getMessage()).getNode();
						msg.setDestination(closest_preceding_finger(aux));
						msg.setNextHopHandle(msg.getDestination());
						try {
							send(msg);
						} catch (QueueFull e) {
							Logger.log("Outgoing Queue of Node " + this.id
									+ " is Full", Logger.ERROR_LOG);
							MessagePool.freeMessage(msg);
						}
					}
					break;
				case GET_PRE :
					//origen --> stabilize();
					msg.setDestination(msg.getSource());
					msg.setNextHopHandle(msg.getSource());
					msg.setMessage(new NodeMessage(predecessor));
					msg.setSource(this.nodeHandle);
					msg.setType(GET_PRE);
					msg.setMode(REPLY);
					try {
						send(msg);
					} catch (QueueFull e) {
						Logger.log("Outgoing Queue of Node " + this.id
								+ " is Full", Logger.ERROR_LOG);
						MessagePool.freeMessage(msg);
					}
					break;
				case SUCC_LIST :
					if (msg.getMode() == REQUEST) {
						SuccListMessage succs = 
							(SuccListMessage) msg.getMessage();
						Vector v = succs.getSuccs();
						if (this.nodeHandle.equals(msg.getSource())) {
							if (v.size() > 0) {
								succ_list = v;
								succ_list.remove(finger[0]);
							}
						} else {
							v.add(this.nodeHandle);
							if (v.size() > ((ChordProperties)Properties.overlayProps).succListMax
									|| finger[0] == null) {
								msg.setDestination(msg.getSource());
								msg.setSource(this.nodeHandle);
								msg.setMode(REPLY);
							} else {
								msg.setDestination(finger[0]);
								msg.setMode(REQUEST);
							}
							msg.setNextHopHandle(msg.getDestination());
							msg.setType(SUCC_LIST);
							succs.setSuccs(v);
							msg.setMessage(succs);
							try {
								send(msg);
							} catch (QueueFull e) {
								Logger.log("Outgoing Queue of Node " + this.id
										+ " is Full", Logger.ERROR_LOG);
								MessagePool.freeMessage(msg);
							}
						}
					} else if (msg.getMode() == REPLY) {
						SuccListMessage succs = 
							(SuccListMessage) msg.getMessage();
						Vector v = succs.getSuccs();
						if (v.size() > 0)
							succ_list = v;
					}
					break;
			}
		}
	}
	/**
	 * Given a time fraction, treats the messages and executes the periodicas
	 * actions, as for example, the stabilization methods
	 * @param actualStep Actual simulation step.
	 * @return true if the node needs to continue the stabilization
	 * process. false if node is just stabilized.
	 */
	public boolean process(int actualStep) {
		clearFingerChanges();
		super.process(actualStep);
		//prettyPrintNode();
		while (hasMoreMessages()) {
			dispatcher(nextMessage());
		}
		if (getFingerChanges()>0) stabRate = 0;
		else if (getFingerChanges() >= realStabilizationRate);
		else stabRate++;
		return stabRate < realStabilizationRate; 
	}
	
	/**
	 * @see planet.commonapi.Node#isAlive() Test if Node is alive
	 */
	public boolean isAlive() {
		return !hasFailed && !hasLeaved;
	}

	public String toString() {
		return "{BadChordNode [" + this.id + "]: succ[" + finger[0] + "]: succ2f["
				+ finger[1] + "]: succList[" + succ_list + "]: pred["
				+ predecessor + "]}";
	}
	
	/**
	 * Return the local NodeHandle.
	 * 
	 * @see planet.commonapi.Node#getLocalHandle()
	 * @return Actual NodeHandle for this Node.
	 */
	public NodeHandle getLocalHandle() {
		return nodeHandle;
	}
	
	
	/* ************************* MESSAGE LISTENERS ********************/
	
	public class LookupListener implements MessageListener {
		private ChordNode node;
		private Id key;
		public LookupListener(ChordNode node, Id key) {
			this.node = node;
			this.key = key;
		}
		public void onMessage(RouteMessage msg) {
			NodeHandle own = ((NodeMessage) msg.getMessage()).getNode();
			Results.addLookup(key, own.getId());
		}
	}
	public class GetPreListener implements MessageListener {
		private ChordNode node;
		public GetPreListener(ChordNode node) {
			this.node = node;
		}
		public void onMessage(RouteMessage msg) {
			// continues stabilize()
			NodeHandle msgNh = ((NodeMessage) msg.getMessage()).getNode();
			if (msgNh != null && msgNh.getId().between(node.getId(), node.getSucc().getId())) {
				node.setSucc(msgNh);
			}
			//send notify
			msg.setDestination(node.getSucc());
			msg.setNextHopHandle(msg.getDestination());
			msg.setSource(node.getLocalHandle());
			msg.setType(NOTIFY);
			msg.setMode(REFRESH);
			msg.setKey(null);
			try {
				node.send(msg);
			} catch (QueueFull e) {
				Logger.log("Outgoing Queue of Node " + node.getId()
						+ " is Full", Logger.ERROR_LOG);
				MessagePool.freeMessage(msg);
			}
		}
	}
	
	
	public class FindSuccListener implements MessageListener {
		private ChordNode node;
		public FindSuccListener(ChordNode node) {
			this.node = node;
		}
		public void onMessage(RouteMessage msg) {
			//continues join()
			NodeHandle succ = ((NodeMessage) msg.getMessage()).getNode();
			node.setSucc(succ);
		}
	}
	
	
	public class FindPredListener implements MessageListener {
		private ChordNode node;
		private java.util.Hashtable info;
		public FindPredListener(ChordNode node, java.util.Hashtable info) {
			this.node = node;
			this.info = info;
		}
		public void onMessage(RouteMessage msg) {
			//n.id from --> n of n.find_successor(n)
			//n1.id to --> n1 of n1.find_successor(n)
			//extract information
			if (info.containsKey("Position")) {
				int pos = ((Integer) info.get("Position")).intValue();
				//n1 executes fix_fingers
				node.setFinger(pos, ((NodeMessage) msg.getMessage()).getNode());
			} else {
				// n1 executes join() --> find_successor(n)
				String key = (String) info.get("Key");
				NodeHandle dest = msg.getDestination();
				msg.setDestination(msg.getSource());
				msg.setNextHopHandle(msg.getDestination());
				msg.setSource(dest);
				msg.setKey(key);
				msg.setType(FIND_SUCC);
				msg.setMode(REPLY);
				try {
					node.send(msg);
				} catch (QueueFull e) {	
					Logger.log("Outgoing Queue of Node " + node.getId()
							+ " is Full", Logger.ERROR_LOG);
					MessagePool.freeMessage(msg);
				}
			}
		}
	}
	
	/**
	 * Sets alive flag to false.
	 * @see planet.commonapi.Node#fail()
	 */
	public void fail() {
		this.hasFailed = true;
		this.nodeHandle.setAlive(false);
	}
	
	/**
	 * Overwrites the outMessages to evaluates the internal flag
	 * <b>hasFailed</b> for not to deliver the outgoing queue of
	 * messages. This implementation is only for Chord overlay.
	 * @see planet.commonapi.Node#outMessages()
	 * @see planet.generic.commonapi.NodeImpl#outMessages()
	 * @return null if the node has failed. The outgoing queue
	 * if the node is alive or just has leaved at this simulation step.
	 */
	public Queue outMessages() {
		if (hasFailed) return null;
		return super.outMessages();
	}

	
	/* END ************************* MESSAGE LISTENERS ********************/
	
	/* ***************** TIMER TASKS *************************************/
	
	/**
	 * Simple TimerTask that invoke stabilize() Node method.
	 * 
	 * @author <a href="mailto: jordi.pujol@estudiants.urv.es">Jordi Pujol </a>
	 *         Date: 07/05/2004
	 */
	public class StabilizeTask extends TimerTaskImpl {
		/**
		 * Initialize this StabilizeTask.
		 */
		public StabilizeTask() {
			super(true);
		}
		/**
		 * Invoke stabilize() method of Node.
		 * 
		 * @see java.util.TimerTask#run()
		 */
		public void run() {
			super.run();
			if (!isFinished()) {
				if (finger[0] != null) {
					stabilize();
				}
			}
		}
		/**
		 * Shows the name of the TimerTask, showing if is periodic.
		 * 
		 * @see java.lang.Object#toString()
		 * @return The name of the TimerTask
		 */
		public String toString() {
			return "StabilizeTask: periodic";
		}
	}
	
	/**
	 * Simple TimerTask that invoke fix_fingers() Node method.
	 * 
	 * @author <a href="mailto: jordi.pujol@estudiants.urv.es">Jordi Pujol </a>
	 *         Date: 07/05/2004
	 */
	public class FixFingerTask extends TimerTaskImpl {
		/**
		 * Initialize this StabilizeTask.
		 */
		public FixFingerTask() {
			super(true);
		}
		/**
		 * Invoke stabilize() method of Node.
		 * 
		 * @see java.util.TimerTask#run()
		 */
		public void run() {
			super.run();
			if (!isFinished()) {
				fix_fingers();
			}
		}
		/**
		 * Shows the name of the TimerTask, showing if is periodic.
		 * 
		 * @see java.lang.Object#toString()
		 * @return The name of the TimerTask
		 */
		public String toString() {
			return "FixFingerTask: periodic";
		}
	}
	
	/* END ************************ TIMER TASKS ****************************/
	
	/* ************************ COMMON API METHODS ****************************/

	/**
	 * Returns a list of nodes that can be used as next hops on a route
	 * towards key. If is safe, the expected fraction of faulty nodes in the
	 * list is guaranteed to be no higher than the fraction of
	 * faulty nodes in the overlay. 
	 * <br><br>
	 * In Chord, must be returns the <b>max</b> successors of <b>key</b> in
	 * the node's location table.
	 * @param key Target key
	 * @param max Number of next hops.
	 * @param safe This flag is not consulted.
	 * @return Until a maximum of max nodes to use as next hop, or null if 
	 * key is not in range of node's location table.
	 * @see planet.commonapi.Node#localLookup(planet.commonapi.Id, int, boolean)
	 */
	public Vector localLookup(Id key, int max, boolean safe) {
		Vector nodes = new Vector(); //to permit only one instance of each one
		int first = firstLocalLookup(key);
		//if key is not in finger table range.
		if (first==-1) return null;
		for (int i=first; i<Properties.bitsKey && nodes.size()<max; i++)
				nodes.add(finger[i]);
		return nodes;
	}
	
	/**
	 * Detects de first position that key is in range of some position of 
	 * finger table.
	 * @param key Id to find the first position of finger table is responsible.
	 * @return -1 if the key is not in range of the finger table, or some value
	 * between 0..Properties.bitsKey-1
	 */
	public int firstLocalLookup(Id key) {
		int i;
		if (key.compareTo(start[Properties.bitsKey-1])>0 ||
			key.compareTo(start[0])<0) return -1;
		for (i=Properties.bitsKey-1;i>=0 && key.compareTo(this.start[i])<=0; i++);
		return i; 
	}
	
	
	/**
	 * @see planet.commonapi.Node#neighborSet(int)
	 * Obtains an unordered list of up to max nodes that are neighbors 
	 * of the local node.
	 * @param max Maximum of nodes to return.
	 * @return Neighbor nodes. At the first position appears the predecessor node,
	 * and following the rest of successors nodes.
	 */
	public Vector neighborSet(int max) {
		int maxIndex = Math.min(max-1,succ_list.size());
		auxCAPI=this.getSuccList(maxIndex);
		auxCAPI.add(0,predecessor);
		return auxCAPI;
	}
	
	/**
	 * @see planet.commonapi.Node#range(planet.commonapi.Id, int, java.lang.Integer, java.lang.Integer)
	 * This operation provides information about ranges of keys for
	 * which the <b>node</b> is currently a root. Returns false if
	 * the range could not be determined, true otherwise.
	 * <br><br>
	 * It is an error to query the range of a node not present in
	 * the neighbor set.
	 * <br><br>
	 * The [leftKey,rightKey] denotes the inclusive range of key values.
	 * @param node Node that is currently a root of some range of keys.
	 * @param rank Number of keys that is root the node (rank=rightKey-leftKey).
	 * @param leftKey The value that appears in the invokation is the candidate left
	 * key of the range. It may be modified to reflect the correct left margin
	 * once invokation has finished.
	 * @param rightkey Shows once the invokation has finished the left margin of
	 * the range. It must be initialized to 0 (zero) before this method could be invoked.
	 * @return true if the range could be determined; false otherwise, including
	 * the case of node is not present in the neighbor set returned by neighborSet().
	 * If false is returned, the leftKey and rightKey are not corrects.
	 */
	public boolean range(NodeHandle node, Id rank, Id leftKey, Id rightKey) {
		/* Must be an error if 'node' is not in array returned by neighborSet()
		 * method. In this case, evaluates the best case of neighborSet()
		 * invokation, using all elements in successor list. */
		if (!node.equals(predecessor) && !succ_list.contains(node)) 
			return false;
		
		try {
			/* Correct the value for 'leftKey'. Only in the case in which the
			 * 'leftkey' is minor than or equals to predecessor for the actual
			 * node. */
			if (leftKey.compareTo(predecessor.getId())<=0) {
				leftKey.add(GenericFactory.buildId(ChordId.ONE_CHORD));
			
			}
		
			/* Sets the value of 'rightKey' to the correct maximum value.
			 * That is, rightKey = min(leftKey+rank,node), once the leftKey 
			 * already has been corrected. */
			rightKey.add(leftKey);
			rightKey.add(rank);
			
			//verify if the sum leaves rank
			BigInteger max = new BigInteger("2").pow(Properties.bitsKey);
			Id maxId = GenericFactory.buildId(max);
			if (maxId.compareTo(rightKey)<=0) {
				rightKey.subtract(maxId);
			}
			//sets the correct value
			if (rightKey.compareTo(node)>0) {
				rightKey.setValue(node);
			}
		} catch (InitializationException e) {
			// because the value cannot be set, return false
			return false;
		}
		return true;
	}
	
	/**
	 * @see planet.commonapi.Node#replicaSet(planet.commonapi.Id, int)
	 * Returns the replica node set.
	 * The number of successor list is defined in the context.
	 * @param key Key from which obtain the replica set.
	 * @param maxRank Number of nodes in replica set.
	 * @return Up to maxRank nodes as replica set.
	 */
	public Vector replicaSet(Id key, int maxRank) {
		if (!key.betweenE(getPred().getId(),getId())) return null;
		if (!(maxRank>=1)) return null;
		auxCAPI = new Vector();
		int maxIndex = Math.min(maxRank-1,succ_list.size());
		if (maxIndex>0) { 
			auxCAPI.addAll(getSuccList(maxIndex));
		}
		auxCAPI.add(0,nodeHandle);
		return auxCAPI;
	}
	
	/* END ************************ COMMON API METHODS ****************************/
	
	
	/**
	 * The EndPoint gets from the node a RouteMessage
	 * initialized with the correct values from type and mode,
	 * for messages builded by application to send a broadcast.
	 * @param appId String that includes the application identification.
	 * @param to Node that must receive this message. If cannot represents
	 * a real node.
	 * @param nextHop Node that must receive this message as first hop.
	 * If it is null, the message must be routed.
	 * @return A RouteMessage with the correctly initialized values.
	 * @throws InitializationException if an error occurs during building
	 * of the RouteMessage.
	 */
/*	public RouteMessage getBroadcastRouteMessage(String appId, NodeHandle to,
			NodeHandle nextHop) throws InitializationException {
		return getBroadcastMessage(appId,nodeHandle,to,nextHop);
	}*/
	
	/**
	 * The EndPoint obtain from the node a RouteMessage 
	 * initialized with the correct values from type and mode,
	 * for messages builded by applications.
	 * @param appId String that includes the application identification.
	 * @param to Node that must receive this message. If cannot represents
	 * a real node.
	 * @param nextHop Node that must receive this message as first hop.
	 * If it is null, the message must be routed.
	 * @return A RouteMessage with the correctly initialized values.
	 * @throws InitializationException if an error occurs during building
	 * of the RouteMessage.
	 */
	/*public RouteMessage getDataRouteMessage(String appId, NodeHandle to,
			NodeHandle nextHop) throws InitializationException {
		return getDataMessage(appId,nodeHandle,to,nextHop);
	}*/
}
