package planet.chord;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import planet.commonapi.exception.InitializationException;
import planet.util.OverlayProperties;

/**
 * This class includes the initialization and the values for all
 * configuration properties of the Chord overlay.
 * @author <a href="mailto: jordi.pujol@estudiants.urv.es">Jordi Pujol</a>
 * Date: 05/07/2004
 */
public class ChordProperties implements OverlayProperties {

	/* *********************** CHORD PROPERTIES ***********************/
	/* Theese must to appear on the properties file */
	/**
	 * Chord property: Default key for number of stabilize steps.
	 */
	public static final String DEFAULT_STABILIZE_STEPS = "DEFAULT_STABILIZE_STEPS";
	/**
	 * Chord property: Default key for number of steps to fix finger tables.
	 */
	public static final String DEFAULT_FIX_FINGER_STEPS = "DEFAULT_FIX_FINGER_STEPS";
	/**
	 * Chord property: Default key for size of successor list.
	 */
	public static final String DEFAULT_SUCC_LIST_MAX = "DEFAULT_SUCC_LIST_MAX";

	
	/* *************** ATTRIBUTES FOR CHORD  ****************/
	/**
	 * Number of steps for statibilization.
	 */
	public int stabilizeSteps;
	/**
	 * Number of steps to fix fingers.
	 */
	public int fixFingerSteps;
	/**
	 * Max size for successor list.
	 */
	public int succListMax;
	
	/**
	 * Initialize all configuration properties of the Chord overlay.
	 * @see planet.util.OverlayProperties#init(java.lang.String)
	 * @param propertiesFile
	 * @throws InitializationException
	 */
	public void init(String propertiesFile) throws InitializationException {
		//Load properties
		FileInputStream fis = null;
		java.util.Properties prop = new java.util.Properties();
		try {
			fis = new FileInputStream(propertiesFile);
			prop.load(fis);
		} catch (FileNotFoundException e) {
			throw new InitializationException("Cannot read the file "+propertiesFile,e);
		} catch (IOException e) {
			throw new InitializationException("Cannot load the file "+propertiesFile,e);
		}
		
		//if all OK
	

		//read the defaultStabilizeSteps
		stabilizeSteps = Integer.parseInt(prop.getProperty(DEFAULT_STABILIZE_STEPS));
		if (stabilizeSteps < 0) 
			throw new InitializationException("Property ["+DEFAULT_STABILIZE_STEPS +"] is not a valid number of stabilize steps.");
		
		//read the defaultFixFingerSteps
		fixFingerSteps = Integer.parseInt(prop.getProperty(DEFAULT_FIX_FINGER_STEPS));
		if (fixFingerSteps < 0) 
			throw new InitializationException("Property ["+DEFAULT_FIX_FINGER_STEPS +"] is not a valid number of steps to fix finger tables.");

		//read the defaultSuccListMax
		succListMax = Integer.parseInt(prop.getProperty(DEFAULT_SUCC_LIST_MAX));
		if (succListMax < 0) 
			throw new InitializationException("Property ["+DEFAULT_SUCC_LIST_MAX +"] is not a valid size of successor list.");
	}
	
	
	
	/**
	 * Returns a String representation of the constant specific values
	 * of type and mode of the RouteMessage. Its use is only for
	 * human readable logs. Based on ChordNode implementation.
	 * @param type Value to get its String representation.
	 * @return The String representation of the type.
	 */
	public String typeToString(int type) {
		return ChordNode.TYPES[type];
	}
}
