package planet.commonapi.factory;

import planet.commonapi.Application;
import planet.commonapi.exception.InitializationException;

/**
 * This interface abstracts the task of generate new instances of Application.
 * In this case, only one parameter exist:
 * <ul>
 * <li><b>The Application's Class</b>: That is, the class of the generate instances
 * of Application, that corresponds to concrete implementation.</li>
 * </ul>
 * There is a default value in the properties file:
 * <ul>
 * <li><b>DEFAULT_APPLICATION</b>: Specifies a concrete implementation of
 * the Application interface.</li>
 * </ul>
 * @author Jordi Pujol
 * @see planet.commonapi.Application Application
 */
public interface ApplicationFactory {
	/**
	 * Generates a new instance of the actual Application class. The name of
	 * Application is setting the default one by its implementation.
	 * @return A new instance of the Application
	 * @see planet.commonapi.Application
	 */
	public Application build() throws InitializationException;
	
	/**
	 * Generates a new instance of the Application class <b>app</b>. This value
	 * must be an existing Application implementation. The Application's name is
	 * setting by the implementation owner.
	 * @param app Class completly defined of the Application to generate only by this method call.
	 * @return A new instance of Application. Its class is <b>app</b>
	 * @see planet.commonapi.Application
	 */
	public Application build(String app) throws InitializationException;
	
	/**
	 * Generates a new instance of the actual Application class. The name of Application is
	 * overwriting by <b>name</b>.
	 * @return A new instance of the Application
	 * @see planet.commonapi.Application
	 * @param name Name for the application to generate.
	 */
	public Application buildWithName(String name) throws InitializationException;
	
	/**
	 * Generates a new instance of the Application class <b>app</b>. This value
	 * must be an existing Application implementation. The name of Application is
	 * overwriting by <b>name</b>.
	 * @param app Class completly defined of the Application to generate only by this method call.
	 * @return A new instance of Application. Its class is <b>app</b>
	 * @see planet.commonapi.Application
	 * @param name Name for the application to generate.
	 */
	public Application buildWithName(String app,String name) throws InitializationException;

	
	/**
	 * Gets the actual Application class whose is generating the new instances of the application,
	 * in calls to the method build()
	 * @return The class of the generated Applications in String format. 
	 * @see build()
	 */
	public String getApplicationClass();
	
	/**
	 * Sets the class of the Application to generate future instances.
	 * @param app Class of the Application in String format.
	 */
	public void setApplicationClass(String app) throws InitializationException;
}
