package planet.commonapi.factory;

import planet.commonapi.Id;
import planet.commonapi.exception.InitializationException;

/**
 * @(#) IdFactory.java This interface provides methods with a way of generating 
 * Ids.  
 * 
 * @author Jordi Pujol
 */
public interface IdFactory extends java.io.Serializable  {
	
	/**
	 * Builds an Id with the actual configuration of network topology and size.
	 * @return A new Id generated with the actual configuration.
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable.
	 */
	public Id buildId() throws InitializationException;
	
	/**
	 * Builds a protocol-specific Id given the source data.
	 * @param material The material to use as double value.
	 * @return A new Id with the double value as internal one.
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable.
	 */
	public Id buildId(int material) throws InitializationException;

	
	/**
	 * Builds a protocol-specific Id given the source data.
	 * @param material The material to use as double value.
	 * @return A new Id with the double value as internal one.
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable.
	 */
	public Id buildId(double material) throws InitializationException;

	
	/**
	 * Builds a protocol-specific Id given the source data.
	 *
	 * @param material The material to use to build the new Id.
	 * @return A new Id with the byte[] value.
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable.
	 */
	public Id buildId(byte[] material) throws InitializationException;

	/**
	 * Builds a protocol-specific Id given the source data as int[].
	 * @param material The internal representation value of the new Id
	 * as int[].
	 * @return A new Id with the int[] value.
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable.
	 */
	public Id buildId(int[] material) throws InitializationException;

	/**
	 * Builds a protocol-specific Id by using the given 
	 * string as source data for a hash function (like SHA-1). The 
	 * returned hash value will be the internal value for the new Id.
	 *
	 * @param materialToHash The string to use as source data for a hash function.
	 * @return The built Id as result to apply a hash function
	 * to the given String.
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable.
	 */
	public Id buildKey(String materialToHash) throws InitializationException; 
	
	/**
	 * Generate an Id from the String as its internal value.
	 * @param material With the String representation of the
	 * internal value.
	 * @return A new Id with the value included in the StringBuffer.  
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable. 
	 */
	public Id buildId(String material) throws InitializationException;
	
	/**
	 * Generate an Id from a BigInteger, that includes its internal value.
	 * @param material BigInteger with the internal value for the new Id.
	 * @return A new Id with the value included in the BigInteger.
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable. 
	 */
	public Id buildId(java.math.BigInteger material) throws InitializationException;

}
