package planet.generic.commonapi.factory;

import java.io.*;
import java.lang.reflect.Constructor;

import planet.util.Properties;
import planet.commonapi.Application;
import planet.commonapi.exception.InitializationException;
import planet.commonapi.factory.*;

/**
 * This factory abstracts the functionality of builds new instances of
 * Application. It is based with only one parameter contained in a 
 * properties file:
 * <ul>
 * <li><b>DEFAULT_APPLICATION</b>: The default application to build.<li>
 * </ul>
 * This property can too be modified by invoking methods, of course.
 * <br><br>
 * The concrete implementation of Application must have two constructors
 * to execute without any problem:
 * <ol>
 * <li><b>Constructor()</b>: A constructor without parameters. In this case,
 * the name of its instances are named by implementation owner.</li>
 * <li><b>Constructor(String)</b>: A constructor with only one parameter. It
 * is the name of the new instance of Application.</li>
 * </ol>
 * @author Jordi Pujol
 */
public class ApplicationFactoryImpl implements ApplicationFactory {
	/**
	 * The default key specified in the properties file that identifies
	 * the class of some Application's implementation.
	 */
	protected static final String DEFAULT_APPLICATION = Properties.DEFAULT_APPLICATION;
	/**
	 * Class for Applications to build.
	 */
	protected Class appClass = null;
	/**
	 * Constructor for applications with some parameter.
	 */
	protected Constructor appConstructor = null;
	/**
	 * Properties file to be read to capture the <b>defaultApplication</b>.
	 */
	protected String propertiesFile = null;
	
	
	/**
	 * This constructor reads the properties' file and initializes
	 * the Class for new Applications. If occur any problem
	 * with this initialization, throws an InitializationException 
	 * exception with short description for the problem.
	 * @throws InitializationException If occur any problem during
	 * initialization steps.
	 * @param propertiesFile Properties' file that contains the
	 * data to be read for this ApplicationFactory.
	 */
	public ApplicationFactoryImpl(String propertiesFile) throws InitializationException {
		this.propertiesFile = propertiesFile;
		//Load properties
		FileInputStream fis = null;
		java.util.Properties appProp = new java.util.Properties();

		try {
			fis = new FileInputStream(propertiesFile);
			appProp.load(fis);
		} catch (FileNotFoundException e) {
			throw new InitializationException("Cannot read the file "+propertiesFile,e);
		} catch (IOException e) {
			throw new InitializationException("Cannot load the file "+propertiesFile,e);
		}
		
		//if all OK
		
		//read defaultApplication
		try {
			Class types[] = {"".getClass()};
			this.appClass = Class.forName(appProp.getProperty(DEFAULT_APPLICATION));
			appConstructor = this.appClass.getConstructor(types);
		} catch (ClassNotFoundException e) {
			throw new InitializationException("Property "+DEFAULT_APPLICATION +" is not a valid class name.",e);
		} catch (SecurityException e) {
			throw new InitializationException("Cannot access to the Constructor "+appClass+"(Id)",e);
		} catch (NoSuchMethodException e) {
			throw new InitializationException("Not exists Constructor "+appClass+"(Id)",e);
		}
	}
	
	/**
	 * This constructor reads the properties' file and initializes
	 * the Class for new Applications. If occur any problem
	 * with this initialization, throws an InitializationException 
	 * exception with short description for the problem.
	 * @throws InitializationException If occur any problem during
	 * initialization steps.
	 * @param propertiesFile Properties' file that contains the
	 * data to be read for this ApplicationFactory.
	 */
	public ApplicationFactoryImpl(Class appClass) throws InitializationException {
		//read defaultApplication
		this.appClass = appClass;
		try {
			Class types[] = {"".getClass()};
			appConstructor = this.appClass.getConstructor(types);
		} catch (SecurityException e) {
			throw new InitializationException("Cannot access to the Constructor "+appClass+"(Id)",e);
		} catch (NoSuchMethodException e) {
			throw new InitializationException("Not exists Constructor "+appClass+"(Id)",e);
		}
	}
		
	/**
	 * Builds a new instance of the actual Application's class with the 
	 * default name.
	 * @see planet.commonapi.factory.ApplicationFactory#build()
	 * @return A new instance of the actual Application's class.
	 */
	public Application build() throws InitializationException {
		try {
			return (Application) appClass.newInstance(); //Constructor without parameters
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+appClass.getName()+"].",e);
		}
	}

	/**
	 * Builds a new instance of the specified Application's class <b>app</b> with the
	 * default name. 
	 * @see planet.commonapi.factory.ApplicationFactory#build(java.lang.String)
	 * @param app Application's class to be generated.
	 * @return A new instance of the specified Application's class.
	 */
	public Application build(String app) throws InitializationException {
		Class appClass;
		try {
			appClass = Class.forName(app);
		} catch (ClassNotFoundException e) {
			throw new InitializationException("The Application classname ["+app+"] not found.",e);
		} catch (SecurityException e) {
			throw new InitializationException("Cannot access to the Constructor "+app+"(Id)",e);
		} 
		
		try {
			return (Application) appClass.newInstance();
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+app+"].",e);
		}
	}
	
	/**
	 * Generates a new instance of the actual class of Application.
	 * @see planet.commonapi.factory.ApplicationFactory#build()
	 * @return A new instance of the actually specified Application class.
	 * @see planet.commonapi.Application Application
	 * @throws InitializationException 
	 * @param name Name for the application to generate.
	 */
	public Application buildWithName(String name) throws InitializationException {
		
		try {
			Object params[] = {name};
			return (Application) appConstructor.newInstance(params); //Constructor without parameters
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+appClass.getName()+"].",e);
		}
	}
	
	/**
	 * Generates a new instance of the specified Application class <b>app</b>.
	 * @see planet.commonapi.factory.ApplicationFactory#build(java.lang.String)
	 * @param app Application class to generate this new instance.
	 * @return A new instance of the Application class <b>app</b>
	 * @param name Name for the application to generate.
	 */
	public Application buildWithName(String app,String name) throws InitializationException {
		
		Class appClass;
		Constructor appConstructor;
		try {
			Class types[] = {"".getClass()};
			appClass = Class.forName(app);
			appConstructor = this.appClass.getConstructor(types);
		} catch (ClassNotFoundException e) {
			throw new InitializationException("The Application classname ["+app+"] not found.",e);
		} catch (SecurityException e) {
			throw new InitializationException("Cannot access to the Constructor "+app+"(Id)",e);
		} catch (NoSuchMethodException e) {
			throw new InitializationException("Not exists Constructor "+app+"(Id)",e);
		}
		try {
			Object params[] = {name};
			return (Application) appConstructor.newInstance(params);
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+app+"].",e);
		}
	}
	
	/**
	 * Gets the Application classname actually in use.
	 * @see planet.commonapi.factory.ApplicationFactory#getApplicationClass()
	 * @return Application classname actually in use.
	 */
	public String getApplicationClass() {
		return appClass.getName();
	}

	/**
	 * Sets the Application classname to use in the future.
	 * @see planet.commonapi.factory.ApplicationFactory#setApplicationClass(java.lang.String)
	 * @param app Application classname for some implementation to use in the future.
	 * @throws InitializationException if exists some problem with the Application classname
	 * <b>app</b>. 
	 */
	public void setApplicationClass(String app) throws InitializationException {
		
		try {
			this.appClass = Class.forName(app);
		} catch (ClassNotFoundException e) {
			throw new InitializationException("The Application classname ["+app+"] not found.",e);
		}
	}
}
