package planet.generic.commonapi.factory;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.Constructor;

import planet.commonapi.Application;
import planet.commonapi.EndPoint;
import planet.commonapi.Node;
import planet.commonapi.exception.InitializationException;
import planet.commonapi.factory.EndPointFactory;
import planet.util.Properties;

/**
 * This class uses the Factory Method pattern design to build new EndPoints.
 * For this, it requires a concrete property at factory properties file:
 * <ul>
 * <li><b>DEFAULT_ENDPOINT</b>: The default EndPoint to build.<li>
 * </ul>
 * Secondly, the implementation class must have a constructor with the
 * following arguments:
 * <ul>
 * <li><b>planet.commonapi.Application</b>: One reference of the Application.<li>
 * <li><b>planet.commonapi.Node</b>: One reference of the underlying Node.<li>
 * </ul>
 * And that's all!!
 * @author Jordi Pujol
 */
public class EndPointFactoryImpl implements EndPointFactory {
	/**
	 * Implementation Class of the EndPoint to use for build new EndPoints.
	 */
	private Class endPointClass = null;
	/**
	 * Constructor of EndPoint implementation class, with two arguments
	 * (planet.commonapi.Application, planet.commonapi.Node).
	 */
	private Constructor endPointConstructor = null;
	
	/**
	 * Constructor for the factory using the specified factory properties file.
	 * @param factoryProperties Properties file with all items for factories.
	 * @throws InitializationException if an error has occurred.
	 */
	public EndPointFactoryImpl(String factoryProperties) throws InitializationException {
		//Load properties
		FileInputStream fis = null;
		java.util.Properties appProp = new java.util.Properties();

		try {
			fis = new FileInputStream(factoryProperties);
			appProp.load(fis);
		} catch (FileNotFoundException e) {
			throw new InitializationException("Cannot read the file "+factoryProperties,e);
		} catch (IOException e) {
			throw new InitializationException("Cannot load the file "+factoryProperties,e);
		}
		
		//if all OK
		
		//read defaultApplication
		try {
			Class types[] = {Application.class,Node.class};
			this.endPointClass = Class.forName(appProp.getProperty(Properties.DEFAULT_END_POINT));
			endPointConstructor = this.endPointClass.getConstructor(types);
		} catch (ClassNotFoundException e) {
			throw new InitializationException("Property "+Properties.DEFAULT_END_POINT +" is not a valid class name.",e);
		} catch (SecurityException e) {
			throw new InitializationException("Cannot access to the Constructor "+endPointClass.getName()+"(Application,Node)",e);
		} catch (NoSuchMethodException e) {
			throw new InitializationException("Not exists Constructor "+endPointClass.getName()+"(Application,Node)",e);
		}

	}
	
	/**
	 * Constructor for the factory using the specified implementation class of EndPoint.
	 * @param endPointClass Implementation class of EndPoint.
	 * @throws InitializationException if an error has occurred.
	 */
	public EndPointFactoryImpl(Class endPointClass) throws InitializationException {
		try {
			Class[] types = {Application.class, Node.class};
			endPointConstructor = endPointClass.getConstructor(types);
		} catch (SecurityException e) {
			throw new InitializationException("Cannot build a constructor for EndPoint class.",e);
		} catch (NoSuchMethodException e) {
			throw new InitializationException("Not exist a constructor with Application and Node types.",e);
		}
	}
	
	/**
	 * Builds a new instance of EndPoint, relating specified Application and Node.
	 * @see planet.commonapi.factory.EndPointFactory#buildEndPoint(planet.commonapi.Application, planet.commonapi.Node)
	 * @param app Application to run over the node.
	 * @param node Node over which run the specified application.
	 */
	public EndPoint buildEndPoint(Application app, Node node) throws InitializationException {
		try {
			return (EndPoint)endPointConstructor.newInstance(new Object[]{app,node});
		} catch (IllegalArgumentException e) {
			throw new InitializationException("Arguments for constructor are not valid",e);
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of EndPoint.",e);
		}
	}
}
