package planet.generic.commonapi.factory;

import planet.commonapi.Application;
import planet.commonapi.EndPoint;
import planet.commonapi.Id;
import planet.commonapi.Network;
import planet.commonapi.Node;
import planet.commonapi.NodeHandle;
import planet.commonapi.exception.InitializationException;
import planet.commonapi.factory.ApplicationFactory;
import planet.commonapi.factory.EndPointFactory;
import planet.commonapi.factory.IdFactory;
import planet.commonapi.factory.NetworkFactory;
import planet.commonapi.factory.NodeFactory;
import planet.commonapi.factory.NodeHandleFactory;
import planet.simulate.Logger;
import planet.simulate.MessagePool;
import planet.util.KeyGen;
import planet.util.Properties;
import planet.util.timer.ControlTimer;

/**
 * It is an abstraction class that follows the Factory Method pattern.
 * That is, it offers different methods to obtain differents instances of
 * factories.
 * @author Jordi Pujol
 */
public class GenericFactory {
	/**
	 * The ApplicationFactory that is used to invoke the application factory methods.
	 */
	private static ApplicationFactory applicationFactory = null;
	/**
	 * The NetworkFactory that is used to invoke the network factory methods.
	 */
	private static NetworkFactory networkFactory         = null;
	/**
	 * The EndPointFactory that is used to invoke the endpoint factory methods.
	 */
	private static EndPointFactory endPointFactory       = null;
	/**
	 * The NodeFactory that is used to invoke the node factory methods.
	 */
	private static NodeFactory nodeFactory               = null;
	/**
	 * The NodeHandleFactory that is used to invoke the nodehandle factory methods.
	 */
	private static NodeHandleFactory nodeHandleFactory   = null;
	/**
	 * The IdFactory that is used to invoke the id factory methods.
	 */
	private static IdFactory idFactory                   = null;
	/**
	 * Properties file to use by default.
	 */
	public static final String PROPERTIES_FILE = "../conf/factories.properties";
	/**
	 * Permits to generate an unique String key for RouteMessages that
	 * can send any Node.
	 */
	private static KeyGen keyGen = null;
	
	/**
	 * Initialize the GenericFactory to load the default properties file. This way
	 * permits return the different factories with the same base properties: the factories
	 * obtained with the methods withot parameters returns the same class of instances.
	 * @throws InitializationException if occurs any problem during initialization.
	 */
	public static void init() throws InitializationException {
		MessagePool.init();
		Logger.setLevel(Properties.logLevel);
		applicationFactory = getApplicationFactory();
		networkFactory     = getNetworkFactory();
		networkFactory.setSerializedFile(Properties.serializedFile);
		endPointFactory    = getEndPointFactory();
		nodeFactory        = getNodeFactory();
		nodeHandleFactory  = getNodeHandleFactory();
		idFactory          = getIdFactory();
		keyGen			   = new KeyGen(0,Integer.MAX_VALUE);
		ControlTimer.init();
		Logger.setLevel(Properties.logLevel);
	}
	
	/**
	 * Builds a new instance of ApplicationFactory. 
	 * @return A new instance of the ApplicationFactory default implementation.
	 * @throws InitializationException if occurs some problem with the factory
	 * initialization.
	 */
	public static ApplicationFactory getApplicationFactory() throws InitializationException {
		return new ApplicationFactoryImpl(Properties.appClass);
	}

	/**
	 * Builds a new instance of NetworkFactory. 
	 * @return A new instance of the NetworkFactory default implementation.
	 * @throws InitializationException if occurs some problem with the factory
	 * initialization.
	 */
	public static NetworkFactory getNetworkFactory() throws InitializationException {
		//Class networkClass, Class nodeFactClass, Class idFactory, Class nodeClass, Class idClass, String topology, int networkSize
		return new NetworkFactoryImpl(Properties.networkClass, Properties.nodeFactClass,Properties.idFactory,Properties.nodeClass,Properties.idClass,Properties.topology,Properties.networkSize);
	}
	
	/**
	 * Builds a new instance of EndPointFactory. 
	 * @return A new instance of the EndPointFactory default implementation.
	 * @throws InitializationException if occurs some problem with the factory
	 * initialization.
	 */
	public static EndPointFactory getEndPointFactory() throws InitializationException {
		return new EndPointFactoryImpl(Properties.endPoint);
	}
	
	/**
	 * Builds a new instance of NodeFactory. 
	 * @return A new instance of the NodeFactory default implementation.
	 * @throws InitializationException if occurs some problem with the factory
	 * initialization.
	 */
	public static NodeFactory getNodeFactory() throws InitializationException {
		return new NodeFactoryImpl(Properties.idFactory, Properties.nodeClass, Properties.idClass,Properties.topology,Properties.networkSize);
	}
	
	/**
	 * Builds a new instance of NodeHandleFactory. 
	 * @return A new instance of the NodeHandleFactory default implementation.
	 * @throws InitializationException if occurs some problem with the factory
	 * initialization.
	 */
	public static NodeHandleFactory getNodeHandleFactory() throws InitializationException {
		return new NodeHandleFactoryImpl(Properties.nodeHandleClass);
	}
	
	/**
	 * Builds a new instance of IdFactory. 
	 * @return A new instance of the IdFactory default implementation.
	 * @throws InitializationException if occurs some problem with the factory
	 * initialization.
	 */
	public static IdFactory getIdFactory() throws InitializationException {
		return new IdFactoryImpl(Properties.idClass,Properties.topology,Properties.networkSize);
	}
	
	/**
	 * Builds a new instance of ApplicationFactory.
	 * @param propertiesFile Properties file from where read the default configuration. 
	 * @return A new instance of the ApplicationFactory default implementation.
	 * @throws InitializationException if occurs some problem with the factory
	 * initialization.
	 */
	public static ApplicationFactory getApplicationFactory(String propertiesFile) throws InitializationException {
		return new ApplicationFactoryImpl(propertiesFile);
	}

	/**
	 * Builds a new instance of NetworkFactory.
	 * @param propertiesFile Properties file from where read the default configuration. 
	 * @return A new instance of the NetworkFactory default implementation.
	 * @throws InitializationException if occurs some problem with the factory
	 * initialization.
	 */
	public static NetworkFactory getNetworkFactory(String propertiesFile) throws InitializationException {
		return new NetworkFactoryImpl(propertiesFile);
	}
	
	/**
	 * Builds a new instance of EndPointFactory. 
	 * @return A new instance of the EndPointFactory default implementation.
	 * @throws InitializationException if occurs some problem with the factory
	 * initialization.
	 */
	public static EndPointFactory getEndPointFactory(String propertiesFile) throws InitializationException {
		return new EndPointFactoryImpl(propertiesFile);
	}
	/**
	 * Builds a new instance of NodeFactory.
	 * @param propertiesFile Properties file from where read the default configuration. 
	 * @return A new instance of the NodeFactory default implementation.
	 * @throws InitializationException if occurs some problem with the factory
	 * initialization.
	 */
	public static NodeFactory getNodeFactory(String propertiesFile) throws InitializationException {
		return new NodeFactoryImpl(propertiesFile);
	}
	
	/**
	 * Builds a new instance of IdFactory.
	 * @param propertiesFile Properties file from where read the default configuration. 
	 * @return A new instance of the IdFactory default implementation.
	 * @throws InitializationException if occurs some problem with the factory
	 * initialization.
	 */
	public static IdFactory getIdFactory(String propertiesFile) throws InitializationException {
		return new IdFactoryImpl(propertiesFile);
	}

	/* **************  ApplicationFactory METHODS *************/
	
	/**
	 * @see planet.commonapi.factory.ApplicationFactory#build()
	 */
	public static Application build() throws InitializationException {
		return applicationFactory.build();
	}

	/**
	 * @see planet.commonapi.factory.ApplicationFactory#build(java.lang.String)
	 */
	public static Application build(String app) throws InitializationException {
		return applicationFactory.build(app);
	}

	/**
	 * @see planet.commonapi.factory.ApplicationFactory#buildWithName(java.lang.String)
	 */
	public static Application buildWithName(String name) throws InitializationException {
		return applicationFactory.buildWithName(name);
	}

	/**
	 * @see planet.commonapi.factory.ApplicationFactory#buildWithName(java.lang.String, java.lang.String)
	 */
	public static Application buildWithName(String app, String name) throws InitializationException {
		return applicationFactory.buildWithName(app,name);
	}

	/**
	 * @see planet.commonapi.factory.ApplicationFactory#getApplicationClass()
	 */
	public static String getApplicationClass() {
		return applicationFactory.getApplicationClass();
	}

	/**
	 * @see planet.commonapi.factory.ApplicationFactory#setApplicationClass(java.lang.String)
	 */
	public static void setApplicationClass(String app) throws InitializationException {
		applicationFactory.setApplicationClass(app);
	}

	/* **************  NetworkFactory METHODS *************/
	
	/**
	 * @see planet.commonapi.factory.NetworkFactory#buildNetwork()
	 */
	public static Network buildNetwork() throws InitializationException {
		return networkFactory.buildNetwork();
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#buildNetwork(int)
	 */
	public static Network buildNetwork(int size) throws InitializationException {
		return networkFactory.buildNetwork(size);
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#buildNetwork(int, planet.commonapi.factory.NodeFactory)
	 */
	public static Network buildNetwork(int size, NodeFactory nodeFactory) throws InitializationException {
		return networkFactory.buildNetwork(size,nodeFactory);
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#buildNetwork(int, java.lang.String)
	 */
	public static Network buildNetwork(int size, String topology) throws InitializationException {
		return networkFactory.buildNetwork(size,topology);
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#buildNetwork(int, planet.commonapi.factory.NodeFactory, java.lang.String)
	 */
	public static Network buildNetwork(int size, NodeFactory nodeFactory, String topology) throws InitializationException {
		return networkFactory.buildNetwork(size,nodeFactory,topology);
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#buildNetwork(int, planet.commonapi.factory.NodeFactory, java.lang.String, int)
	 */
	public static Network buildNetwork(int size, NodeFactory nodeFactory, String topology, int k) throws InitializationException {
		return networkFactory.buildNetwork(size,nodeFactory,topology,k);
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#setNodeFactory(planet.commonapi.factory.NodeFactory)
	 */
	public static void setNodeFactory(NodeFactory nodeFactory) {
		networkFactory.setNodeFactory(nodeFactory);
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#getNetworkSize()
	 */
	public static int getNetworkSize() {
		return networkFactory.getNetworkSize();
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#setNetworkSize(int)
	 */
	public static void setNetworkSize(int size) {
		networkFactory.setNetworkSize(size);
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#getTopology()
	 */
	public static String getTopology() {
		return networkFactory.getTopology();
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#setTopology(java.lang.String)
	 */
	public static void setTopology(String topology) throws InitializationException {
		networkFactory.setTopology(topology);
		
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#getSerializedFile()
	 */
	public static String getSerializedFile() {
		return networkFactory.getSerializedFile();
	}

	/**
	 * @see planet.commonapi.factory.NetworkFactory#setSerializedFile(java.lang.String)
	 */
	public static void setSerializedFile(String file) throws InitializationException {
		networkFactory.setSerializedFile(file);
		
	}
	
	/* **************  EndPointFactory METHODS *************/
	
	/**
	 * @see planet.commonapi.factory.EndPointFactory#buildEndPoint(planet.commonapi.Application,planet.commonapi.Node)
	 */
	public static EndPoint buildEndPoint(Application app, Node node) throws InitializationException {
		return endPointFactory.buildEndPoint(app,node);
	}
	
	/* **************  NodeFactory METHODS *************/

	/**
	 * @see planet.commonapi.factory.NodeFactory#buildNode()
	 */
	public static Node buildNode() throws InitializationException {
		return nodeFactory.buildNode();
	}

	/**
	 * @see planet.commonapi.factory.NodeFactory#buildNode(planet.commonapi.Id)
	 */
	public static Node buildNode(Id id) throws InitializationException {
		return nodeFactory.buildNode(id);
	}

	/**
	 * @see planet.commonapi.factory.NodeFactory#setIdFactory(planet.commonapi.factory.IdFactory)
	 */
	public static void setIdFactory(IdFactory idFactory) {
		nodeFactory.setIdFactory(idFactory);
		
	}

	/**
	 * @see planet.commonapi.factory.NodeFactory#getNodeClass()
	 */
	public static Class getNodeClass() {
		return nodeFactory.getNodeClass();
	}

	/**
	 * @see planet.commonapi.factory.NodeFactory#setNodeClass(java.lang.Class)
	 */
	public static void setNodeClass(Class nodeClass) throws InitializationException {
		nodeFactory.setNodeClass(nodeClass);
	}

	/* **************  IdFactory METHODS *************/
	
	/**
	 * @see planet.commonapi.factory.IdFactory#buildId()
	 */
	public static Id buildId() throws InitializationException {
		return idFactory.buildId();
	}
	
	/**
	 * @see planet.commonapi.factory.IdFactory#buildId(int)
	 */
	public static Id buildId(int material) throws InitializationException {
		return idFactory.buildId(material);
	}
	
	/**
	 * @see planet.commonapi.factory.IdFactory#buildId(double)
	 */
	public static Id buildId(double material) throws InitializationException {
		return idFactory.buildId(material);
	}

	/**
	 * @see planet.commonapi.factory.IdFactory#buildId(byte[])
	 */
	public static Id buildId(byte[] material) throws InitializationException {
		return idFactory.buildId(material);
	}

	/**
	 * @see planet.commonapi.factory.IdFactory#buildId(int[])
	 */
	public static Id buildId(int[] material) throws InitializationException {
		return idFactory.buildId(material);
	}

	/**
	 * @see planet.commonapi.factory.IdFactory#buildKey(java.lang.String)
	 */
	public static Id buildKey(String string) throws InitializationException {
		return idFactory.buildKey(string);
	}
	
	/**
	 * @see planet.commonapi.factory.IdFactory#buildId(java.lang.String)
	 */
	public static Id buildId(String string) throws InitializationException {
		return idFactory.buildId(string);
	}

	/**
	 * @see planet.commonapi.factory.IdFactory#buildId(java.math.BigInteger)
	 */
	public static Id buildId(java.math.BigInteger bigNumber) throws InitializationException {
		return idFactory.buildId(bigNumber);
	}

	/**
	 * Generate a unique key, based on a simple int. Based on the context of
	 * the simulator, its implementation ensures that the returned key is
	 * unique at the time that it is generated. This key is used to identify
	 * a unique RouteMessage.
	 * @return An unique key for a RouteMessage to be send.
	 */
	public static String generateKey() {
		return keyGen.generateKey();
	}
	
	/* **************  NodeHandleFactory METHODS *************/
	
	/**
	 * Builds a new NodeHandle for specifieds Node Id and alive flag.
	 * @see planet.commonapi.factory.NodeHandleFactory#buildNodeHandle(planet.commonapi.Id, boolean)
	 * @param nodeId Id of the related Node.
	 * @param alive true if the related Node is alive. false in other case.
	 * @return A new instance of NodeHandle for the related Node.
	 * @throws InitializationException if occur any problem.
	 */
	public static NodeHandle buildNodeHandle(Id nodeId, boolean alive)
			throws InitializationException {
		return nodeHandleFactory.buildNodeHandle(nodeId,alive);
	}
	/**
	 * Gets the NodeHandle class actually in use to build new NodeHandles.
	 * @see planet.commonapi.factory.NodeHandleFactory#getNodeHandleClass()
	 * @return The NodeHandle class actually in use.
	 */
	public static Class getNodeHandleClass() {
		return nodeHandleFactory.getNodeHandleClass();
	}
	
	/** 
	 * Sets the NodeHandle class to use in the future.
	 * @see planet.commonapi.factory.NodeHandleFactory#setNodeHandleClass(java.lang.Class)
	 * @param nodeHandleClass NodeHandle class actually in use.
	 * @throws InitializationException if any error has ocurred.
	 */
	public void setNodeHandleClass(Class nodeHandleClass)
			throws InitializationException {
		nodeHandleFactory.setNodeHandleClass(nodeHandleClass);
	}
}
