package planet.generic.commonapi.factory;

import java.io.IOException;
import java.math.BigInteger;

import planet.commonapi.Id;
import planet.commonapi.exception.InitializationException;

/**
 * This Factory generate uniques Id from a material. The Id must to be
 * based on a double representation of its internal value.
 * @author Jordi Pujol
 */
public class IdFactoryDoubleImpl extends IdFactoryImpl {
	
	/**
	 * This constructor reads the properties' file and initializes
	 * the IdFactory and Class for new Nodes. If occur any problem
	 * with this initialization, throws a InitializationException 
	 * exception with short description for the problem.
	 * @throws InitializationException If occur any problem during
	 * initialization steps.
	 */
	public IdFactoryDoubleImpl(String propertiesFile) throws InitializationException {
		super(propertiesFile);
	}
	
	/**
	 * Constructor that builds an instance with these parameters
	 * @param idClass Id class to build.
	 * @param topology Network topology to known how build the new Ids.
	 * @param networkSize Size of the network. It is important if is used the topology Topology.CIRCULAR.
	 * @throws InitializationException if occurs some error during initialization of values.
	 * @see planet.generic.commonapi.factory.Topology Topology
	 */
	public IdFactoryDoubleImpl(Class idClass,String topology,int networkSize) throws InitializationException {
		super(idClass,topology,networkSize);
	}
	
	/**
	 * Initializes the protected attributes to permits the uniform
	 * distribution of Ids. 
	 * @throws InitializationException if an error occurs during
	 * their initialization.
	 */
	protected void initDistributedAttr() throws InitializationException {
		if (this.networkSize != 0) {
			//initialize actualValue and chunkValue to generate distributed Ids (for CIRCULARs networks)
			double MAX = 1.0; //maximum
			double res = MAX/(double)(this.networkSize+1); //value to add to the initial value (0)
			
			//generate the chunkValue
			Object params[] = {new Double(res)};
			try {
				chunkValue = (Id) idClassConstructorDouble.newInstance(params);
			} catch (Exception e) {
				throw new InitializationException("The chunk value for Ids cannot be initialized.",e);
			}
			
			//generate the actualValue (-chunkValue)
			try{
				actualValue = (Id) idClass.newInstance();
				actualValue = actualValue.subtract(chunkValue);
			} catch (Exception e) {
				throw new InitializationException("The initial value for the first Id cannot be initialized.",e);
			}
		}
		//sets the actual number of Ids generated (0)
		this.actualBuildsIds = 0;
	}
	
	
	/**
	 * Builds a random Id using the double constructor of the target Id.
	 * @return A new Id builded randomly.
	 * @throws InitializationException if an error occurs during the
	 * initialization of the Id.
	 */
	protected Id buildRandomId() throws InitializationException {
		Id toReturn = null;
		Object params[] = {new Double(random.nextDouble())};
		try {
			toReturn = (Id) idClassConstructorDouble.newInstance(params);
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+ idClass.getName()
					+"] with this BigInteger ["+ params[0].toString() +"].",e);
		}
		return toReturn;
	}
	
	/**
	 * Builds an Id with the double parameter as its internal value.
	 * @see planet.commonapi.factory.IdFactory#buildId(double)
	 * @param material double with the internal value for the new Id.
	 * @return New Id with the double internal value.
	 * @throws InitializationException if an error occurs during the
	 * initialization of the Id or if it is nonapplicable to the 
	 * target Id.
	 */
	public Id buildId(double material) throws InitializationException {
		Object params[] = {new Double(material)};
		try {
			return (Id) idClassConstructorDouble.newInstance(params);
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+ idClass.getName() 
					+"] with this double ["+ material +"].",e);
		}
	}
	
	/**
	 * Builds an Id with the int parameter as its internal value.
	 * @see planet.commonapi.factory.IdFactory#buildId(int)
	 * @param material Int with the internal value for the new Id.
	 * @return New Id with the int internal value.
	 * @throws Always InitializationException for nonapplicable constructor.
	 */
	public Id buildId(int material) throws InitializationException {
		throw new InitializationException("ERROR: It is not implemented.");
	}
	
	/**
	 * Generate an Id from material in byte[] format as its internal value.
	 * @param material Hash code previously generated.
	 * @see planet.commonapi.factory.IdFactory#buildId(byte[])
	 * @throws InitializationException if an error occurs during the
	 * initialization of the Id or if it is nonapplicable to the 
	 * target Id.
	 */
	public Id buildId(byte[] material) throws InitializationException {
		Object params[] = {material};
		try {
			return (Id) idClassConstructorByteArray.newInstance(params);
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+ idClass.getName() 
					+"] with this byte[] ["+ material +"].",e);
		}
	}

	/**
	 * Generate an Id from material in int[] format, as its internal value.
	 * @param material Internal value of the new Id in int[] format.
	 * @see planet.commonapi.factory.IdFactory#buildId(int[])
	 * @throws InitializationException if an error occurs during the
	 * initialization of the Id or if it is nonapplicable to the 
	 * target Id.
	 */
	public Id buildId(int[] material) throws InitializationException {
		throw new InitializationException("ERROR: It is not implemented.");
	}

	/**
	 * Generate an Id from the BigInteger as its internal value.
	 * @param biNumber BigInteger with the internal value of the new Id. 
	 * @see planet.commonapi.factory.IdFactory#buildId(java.math.BigInteger)
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable. 
	 */
	public Id buildId(BigInteger bigNumber) throws InitializationException {
		throw new InitializationException("ERROR: It is not implemented.");
	}
	
	/**
	 * Makes nothing special, only invokes to <b>stream.defaultWriteObject()</b>.
	 * @param stream Stream to save the actual instance.
	 * @throws IOException if any error has ocurred.
	 */
	private void writeObject(java.io.ObjectOutputStream stream)
    throws IOException {
		writeIdFactoryObject(stream);
	}

	/**
	 * Makes a first <b>stream.defaultReadObject()</b> invocation to uses
	 * the default read method. And then, builds the constructor for
	 * the loaded Node class.
	 * @param stream Stream to read the actual instance.
	 * @throws IOException if occur any error during 
	 * default read object, or if there are any error during
	 * building of Node constructor.  
	 * @throws ClassNotFoundException
	 */
	private void readObject(java.io.ObjectInputStream stream)
    throws IOException, ClassNotFoundException {
		readIdFactoryObject(stream);
	}
}
