package planet.generic.commonapi.factory;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.lang.reflect.Constructor;
import java.math.BigInteger;
import java.util.Properties;
import java.util.Random;

import planet.commonapi.Id;
import planet.commonapi.exception.InitializationException;
import planet.commonapi.factory.IdFactory;
import planet.util.Utilities;

/**
 * This Factory generate uniques Id from a material.
 * @author Jordi Pujol
 */
public class IdFactoryImpl implements IdFactory {
	/**
	 * Properties file name whose read properties.
	 */
	protected String propertiesFile = "";
	
	
	/* ********* PROPERTIES INTO THE CONFIGURATION FILE *************/
	
	/**
	 * The default Class for Ids specified in the properties file.
	 */
	protected static final String defaultIdClass = "DEFAULT_IDCLASS";
	/**
	 * The default topology of the target network specified in the properties file.
	 */
	protected static final String defaultTopology = "DEFAULT_TOPOLOGY";
	/**
	 * The default size key for the target network specified in the properties file.
	 */
	protected static final String defaultNetworkSize = "DEFAULT_SIZE";
	
	/* END ********* PROPERTIES INTO THE CONFIGURATION FILE *************/
	
	/* ************************* CONSTRUCTORS OF ID ************************/
	
	/**
	 * Constructor of actual idClass with only one parameter: an int.
	 */
	protected transient Constructor idClassConstructorInt = null;
	/**
	 * Constructor of actual idClass with only one parameter: a double.
	 */
	protected transient Constructor idClassConstructorDouble = null;
	/**
	 * Constructor of actual idClass with only one parameter: a byte array.
	 */
	protected transient Constructor idClassConstructorByteArray = null;
	/**
	 * Constructor of actual idClass with only one parameter: an int array.
	 */
	protected transient Constructor idClassConstructorIntArray = null;
	/**
	 * Constructor of actual idClass with only one parameter: a String.
	 */
	protected transient Constructor idClassConstructorString = null;
	/**
	 * Constructor of actual idClass with only one parameter: a BigInteger.
	 */
	protected transient Constructor idClassConstructorBigInteger = null;
	
	/* END *********************** CONSTRUCTORS OF ID ************************/
	
	/* ************************* VALUES OF PROPERTIES ************************/
	/**
	 * Actual instance of Class for new Ids.
	 */
	protected Class idClass = null;
	/**
	 * Topology of the actual target network. It is necessary to know how
	 * generate the new Ids of Nodes.
	 */
	protected String topology = null;
	/**
	 * Network size for the target network.
	 */
	protected int networkSize;
	
	/* END ************************* VALUES OF PROPERTIES ********************/
	
	/* *********************** ATTRIBUTES TO BUILD DISTRIBUTED ID ************/
	
	/**
	 * Id just generated. 
	 */
	protected Id actualValue = null;
	/**
	 * Actual value to add to the <b>actualValue</b> to obtain the next Id.
	 */
	protected Id chunkValue = null;
	/**
	 * Number of Ids generated actually. Its value is important for the
	 * Circular networks.
	 */
	protected int actualBuildsIds;
	
	/* END ******************* ATTRIBUTES TO BUILD DISTRIBUTED ID ************/
	
	/**
	 * Random generator for random Id's. 
	 */
	protected Random random = null;
	
	
	/**
	 * This constructor reads the properties' file and initializes
	 * the IdFactory and Class for new Nodes. If occur any problem
	 * with this initialization, throws a InitializationException 
	 * exception with short description for the problem.
	 * @throws InitializationException If occur any problem during
	 * initialization steps.
	 */
	public IdFactoryImpl(String propertiesFile) throws InitializationException {
		this.propertiesFile = propertiesFile;
		//Load properties
		FileInputStream fis = null;
		Properties idProp = new Properties();
		String propertiesFault = "";
		try {
			fis = new FileInputStream(propertiesFile);
			idProp.load(fis);
		} catch (FileNotFoundException e) {
			throw new InitializationException("Cannot read the file "+propertiesFile,e);
		} catch (IOException e) {
			throw new InitializationException("Cannot load the file "+propertiesFile,e);
		}
		
		//if all OK
		
		//read defaultIdClass
		try {
			idClass = Class.forName(idProp.getProperty(defaultIdClass));
		} catch (ClassNotFoundException e) {
			throw new InitializationException("Property ["+defaultIdClass +"] is not a valid class name.",e);
		} catch (Exception e) {
			throw new InitializationException("Cannot obtain a constructor for ["+ idClass.getName() +"].",e);
		}
		
		init(idClass,idProp.getProperty(defaultTopology),Integer.parseInt(idProp.getProperty(defaultNetworkSize)));
	}
	
	/**
	 * Constructor that builds an instance with these parameters
	 * @param idClass Id class to build.
	 * @param topology Network topology to known how build the new Ids.
	 * @param networkSize Size of the network. It is important if is used the topology Topology.CIRCULAR.
	 * @throws InitializationException if occurs some error during initialization of values.
	 * @see planet.generic.commonapi.factory.Topology Topology
	 */
	public IdFactoryImpl(Class idClass,String topology,int networkSize) throws InitializationException {
		init(idClass,topology,networkSize);
	}
	
	/**
	 * Initialize the new instance with correct values to generate new Ids.
	 * @param idClass Id class to be build with this factory.
	 * @param topology Network topology of the target network.
	 * @param networkSize Number of nodes to be build to the target network.
	 * @throws InitializationException if occurs some error with the initialization process.
	 */
	protected void init(Class idClass,String topology,int networkSize) throws InitializationException {
		//get constructors for idClass
		this.idClass = idClass;
		try {
			Class types[]  = {int.class};
			Class types2[] = {double.class};
			Class types3[] = {byte[].class};
			Class types4[] = {int[].class}; 
			Class types5[] = {String.class};
			Class types6[] = {BigInteger.class};
			idClassConstructorInt          = idClass.getConstructor(types);
			idClassConstructorDouble       = idClass.getConstructor(types2);
			idClassConstructorByteArray    = idClass.getConstructor(types3);
			idClassConstructorIntArray     = idClass.getConstructor(types4);
			idClassConstructorString       = idClass.getConstructor(types5);
			idClassConstructorBigInteger   = idClass.getConstructor(types6);
		} catch (Exception e) {
			throw new InitializationException("Cannot obtain constructors for ["+ idClass.getName() +"].",e);
		} 
		
		//read defaultTopology of network
		this.topology = topology;
		if (!Topology.isValid(topology))
			throw new InitializationException("The topology ["+topology+"] is not a valid topology to build new Ids.");
		
		//initialize Random generator
		this.random = new Random();
		
		//read the defaultSize of network.
		this.networkSize = networkSize;
		if (this.networkSize < 0) {
			throw new InitializationException("The network size ["+networkSize+"] are invalid.");
		}
		
		initDistributedAttr();
	}
	
	/**
	 * Initializes the protected attributes to permits the uniform
	 * distribution of Ids. 
	 * @throws InitializationException if an error occurs during
	 * their initialization.
	 */
	protected void initDistributedAttr() throws InitializationException {
		if (this.networkSize != 0) {
			//initialize actualValue and chunkValue to generate distributed Ids (for CIRCULARs networks)
			BigInteger MAX = new BigInteger("2").pow(planet.util.Properties.bitsKey); //maximum
			BigInteger res = MAX.divide(BigInteger.valueOf((long)(this.networkSize+1))); //value to add to the initial value (0)
			
			//generate the chunkValue
			Object params[] = {res.toString()};
			try {
				chunkValue = (Id) idClassConstructorString.newInstance(params);
			} catch (Exception e) {
				throw new InitializationException("The chunk value for Ids cannot be initialized.",e);
			}
			
			//generate the actualValue (-chunkValue)
			try{
				actualValue = (Id) idClass.newInstance();
				actualValue = actualValue.subtract(chunkValue);
			} catch (Exception e) {
				throw new InitializationException("The initial value for the first Id cannot be initialized.",e);
			}
		}
		//sets the actual number of Ids generated (0)
		this.actualBuildsIds = 0;
	}
	
	
	/**
	 * Builds an Id with the actual configuration of network topology and size.
	 * Use the protected method buildRandomId() to build the Id if the
	 * specified topology is random.
	 * @return A new Id generated with the actual configuration.
	 */
	public Id buildId() throws InitializationException {
		Id toReturn = null;
		if (this.topology.equals(Topology.RANDOM)) {
			toReturn = buildRandomId();
		} else if (this.topology.equals(Topology.CIRCULAR)) {
			if (this.actualBuildsIds >= this.networkSize ) {
				throw new InitializationException("Cannot build a new instance of ["+ idClass.getName()
						+"]. The topology network is ["+ Topology.CIRCULAR +"] and just generated all possible Ids ["+
						this.networkSize+"].");
			}
			this.actualValue = this.actualValue.add(this.chunkValue);
			this.actualBuildsIds++;
			toReturn = actualValue;
		} else {
			throw new InitializationException("Cannot build a new Id below the actual network topology ["+
					this.topology +"].");
		}
		return toReturn;
	}
	
	/**
	 * Builds a random Id using the BigInteger constructor of the target Id.
	 * @return A new Id builded randomly. This method must be reimplemented
	 * by the IdFactories that have another requirements to build them.
	 * @throws InitializationException if an error occurs during the
	 * initialization of the Id.
	 */
	protected Id buildRandomId() throws InitializationException {
		Id toReturn = null;
		Object params[] = {new BigInteger(planet.util.Properties.bitsKey,random)};
		try {
			toReturn = (Id) idClassConstructorBigInteger.newInstance(params);
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+ idClass.getName()
					+"] with this BigInteger ["+ params[0].toString() +"].",e);
		}
		return toReturn;
	}
	
	/**
	 * Builds an Id with the double parameter as its internal value.
	 * @see planet.commonapi.factory.IdFactory#buildId(double)
	 * @param material double with the internal value for the new Id.
	 * @return New Id with the double internal value.
	 * @throws InitializationException if an error occurs during the
	 * initialization of the Id or if it is nonapplicable to the 
	 * target Id.
	 */
	public Id buildId(double material) throws InitializationException {
		throw new InitializationException("ERROR: It is not implemented.");
	}
	
	/**
	 * Builds an Id with the int parameter as its internal value.
	 * @see planet.commonapi.factory.IdFactory#buildId(int)
	 * @param material Int with the internal value for the new Id.
	 * @return New Id with the int internal value.
	 * @throws InitializationException if an error occurs during the
	 * initialization of the Id or if it is nonapplicable to the 
	 * target Id.
	 */
	public Id buildId(int material) throws InitializationException {
		throw new InitializationException("ERROR: It is not implemented.");
	}
	
	/**
	 * Generate an Id from material in byte[] format as its internal value.
	 * @param material Hash code previously generated.
	 * @see planet.commonapi.factory.IdFactory#buildId(byte[])
	 * @throws InitializationException if an error occurs during the
	 * initialization of the Id or if it is nonapplicable to the 
	 * target Id.
	 */
	public Id buildId(byte[] material) throws InitializationException {
		Object params[] = {material};
		try {
			return (Id) idClassConstructorByteArray.newInstance(params);
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+ idClass.getName() 
					+"] with this byte[] ["+ material +"].",e);
		}
	}

	/**
	 * Generate an Id from material in int[] format, as its internal value.
	 * @param material Internal value of the new Id in int[] format.
	 * @see planet.commonapi.factory.IdFactory#buildId(int[])
	 * @throws InitializationException if an error occurs during the
	 * initialization of the Id or if it is nonapplicable to the 
	 * target Id.
	 */
	public Id buildId(int[] material) throws InitializationException {
		Object params[] = {material};
		try {
			return (Id) idClassConstructorIntArray.newInstance(params);
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+ idClass.getName() 
					+"] with this int[] ["+ material +"].",e);
		}
	}

	/**
	 * Generate an Id from a String with SHA-1 hash function.
	 * This method use the constructor of the implemented Id with 
	 * byte[] argument.
	 * <br><br>
	 * The implementation makes the following operation:
	 * <br>
	 * <center><b>Utilities.generateByteHash(string)</b></center>
	 * <br>
	 * to build a byte[] with its hash value.
	 * @param string String to apply default hash function (SHA-1) to 
	 * generate Id. 
	 * @see planet.commonapi.factory.IdFactory#buildKey(java.lang.String)
	 * @see planet.util.Utilities.generateByteHash(java.lang.String)
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable. 
	 */
	public Id buildKey(String string) throws InitializationException {
		
		Object params[] = {Utilities.generateByteHash(string)};
		try {
			return (Id) idClassConstructorByteArray.newInstance(params);
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance " +
					"of ["+ idClass.getName() +"] with this String ["+ string +
					"].",e);
		}
	}
	
	/**
	 * Generate an Id from a String that contains its internal value.
	 * @param string Contains the internal value of Id in String format. 
	 * @see planet.commonapi.factory.IdFactory#buildId(java.lang.String)
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable. 
	 */
	public Id buildId(String material) throws InitializationException {
		Object params[] = {material};
		try {
			return (Id) idClassConstructorString.newInstance(params);
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+ idClass.getName() 
					+"] with this StringBuffer ["+ material +".",e);
		}
	}
	
	/**
	 * Generate an Id from the BigInteger as its internal value.
	 * @param biNumber BigInteger with the internal value of the new Id. 
	 * @see planet.commonapi.factory.IdFactory#buildId(java.math.BigInteger)
	 * @throws InitializationException when an error occurs during the
	 * initialization of the Id or when this factory method is
	 * nonapplicable. 
	 */
	public Id buildId(BigInteger bigNumber) throws InitializationException {
		Object params[] = {bigNumber};
		try {
			return (Id) idClassConstructorBigInteger.newInstance(params);
		} catch (Exception e) {
			throw new InitializationException("Cannot build a new instance of ["+ idClass.getName() 
					+"] with this BigInteger ["+ bigNumber +".",e);
		}
	}
	
	/**
	 * Makes nothing special, only invokes to <b>stream.defaultWriteObject()</b>.
	 * @param stream Stream to save the actual instance.
	 * @throws IOException if any error has ocurred.
	 */
	private void writeObject(java.io.ObjectOutputStream stream)
    throws IOException {
		writeIdFactoryObject(stream);
	}
	
	/**
	 * Makes nothing special, only invokes to <b>stream.defaultWriteObject()</b>.
	 * This must be invoked by the different writeObject() method.
	 * @param stream Stream to save the actual instance.
	 * @throws IOException if any error has ocurred.
	 */
	protected void writeIdFactoryObject(java.io.ObjectOutputStream stream)
    throws IOException {
		stream.defaultWriteObject();
	}

	/**
	 * Makes a first <b>stream.defaultReadObject()</b> invocation to uses
	 * the default read method. And then, builds the constructor for
	 * the loaded Node class.
	 * @param stream Stream to read the actual instance.
	 * @throws IOException if occur any error during 
	 * default read object, or if there are any error during
	 * building of Node constructor.  
	 * @throws ClassNotFoundException
	 */
	private void readObject(java.io.ObjectInputStream stream)
    throws IOException, ClassNotFoundException {
		readIdFactoryObject(stream);
	}
	
	/**
	 * This method must be invoked by the different IdFactory
	 * implementation to recover all Constructors that this class
	 * contains.
	 * @param stream Stream to read the actual instance.
	 * @throws IOException if occur any error during 
	 * default read object, or if there are any error during
	 * building of Node constructor.  
	 * @throws ClassNotFoundException
	 */
	protected void readIdFactoryObject(ObjectInputStream stream) 
			throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		try {
			Class types[]  = {int.class};
			Class types2[] = {double.class};
			Class types3[] = {byte[].class};
			Class types4[] = {int[].class}; 
			Class types5[] = {String.class};
			Class types6[] = {BigInteger.class};
			idClassConstructorInt          = idClass.getConstructor(types);
			idClassConstructorDouble       = idClass.getConstructor(types2);
			idClassConstructorByteArray    = idClass.getConstructor(types3);
			idClassConstructorIntArray     = idClass.getConstructor(types4);
			idClassConstructorString       = idClass.getConstructor(types5);
			idClassConstructorBigInteger   = idClass.getConstructor(types6);
		} catch (Exception e) {
			throw new IOException("Cannot obtain constructors for ["+ idClass.getName() +"].");
		}
	}
}
