package planet.generic.commonapi.factory;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.Constructor;


import planet.util.Properties;
import planet.commonapi.Id;
import planet.commonapi.Node;
import planet.commonapi.exception.InitializationException;
import planet.commonapi.factory.*;

/**
 * This is a specific implementation of NodeFactory that permits
 * build any class of Node with any type of Id. It is possible by
 * identifies theese parameters in properties file:
 * <ol>
 * <li><b>DEFAULT_IDFACTORY</b>: Class which identifies the IdFactory to use.</li> 
 * <li><b>DEFAULT_NODECLASS</b>: Class which identifies the Class for new building instances.</li>
 * <li><b>DEFAULT_TOPOLOGY</b>: Topology of the network to construct. It is
 * necessary to construct a valid Id for the new Node.</li>
 * </ol>
 * They can be modified too by invoking methods to this class. In any case, 
 * the properties file only can be modified by programmer.
 * @author Jordi Pujol
 */
public class NodeFactoryImpl implements NodeFactory {
	/**
	 * Properties file name whose read properties.
	 */
	protected String propertiesFile = "";
	/**
	 * Actual instance of IdFactory.
	 */
	private IdFactory idFactory = null;
	/**
	 * Actual instance of Class for new Nodes.
	 */
	private Class nodeClass = null;
	/**
	 * Constructor of actual nodeClass with only one parameter: an Id object.
	 */
	private transient Constructor nodeClassConstructor = null;
	/**
	 * Topology of the actual target network. It is necessary to know how
	 * generate the new Ids of Nodes.
	 */
	private String topology = null;
	
	
	/**
	 * This constructor reads the properties' file and initializes
	 * the IdFactory and Class for new Nodes. If occur any problem
	 * with this initialization, throws a InitializationException 
	 * exception with short description for the problem.
	 * @throws InitializationException If occur any problem during
	 * initialization steps.
	 */
	public NodeFactoryImpl(String propertiesFile) throws InitializationException {
		this.propertiesFile = propertiesFile;
		//Load properties
		FileInputStream fis = null;
		java.util.Properties nodeProp = new java.util.Properties();
		String propertiesFault = "";
		try {
			fis = new FileInputStream(propertiesFile);
			nodeProp.load(fis);
		} catch (FileNotFoundException e) {
			throw new InitializationException("Cannot read the file "+propertiesFile,e);
		} catch (IOException e) {
			throw new InitializationException("Cannot load the file "+propertiesFile,e);
		}
		
		//if all OK
		
		//read Properties.DEFAULT_IDFACTORY
		try {
			Class idFactClass = Class.forName(nodeProp.getProperty(Properties.DEFAULT_IDFACTORY));
			Class[] types = {String.class};
			Constructor construct = idFactClass.getConstructor(types);
			Object[] params = {propertiesFile};
			idFactory = (IdFactory) construct.newInstance(params);
		} catch (ClassNotFoundException e) {
			throw new InitializationException("Property "+Properties.DEFAULT_IDFACTORY +" is not a valid class name.",e);
		} catch (Exception e) {
			throw new InitializationException("Cannot generate an instance of IdFactory.",e);
		} 
		
		//read Properties.DEFAULT_NODECLASS
		try {
			Class types[] = {Id.class};
			this.nodeClass = Class.forName(nodeProp.getProperty(Properties.DEFAULT_NODECLASS));
			nodeClassConstructor = this.nodeClass.getConstructor(types);
		} catch (ClassNotFoundException e) {
			throw new InitializationException("Property "+Properties.DEFAULT_NODECLASS +" is not a valid class name.",e);
		} catch (SecurityException e) {
			throw new InitializationException("Cannot access to the Constructor "+nodeClass+"(Id)",e);
		} catch (NoSuchMethodException e) {
			throw new InitializationException("Not exists Constructor "+nodeClass+"(Id)",e);
		}
		this.topology = nodeProp.getProperty(Properties.DEFAULT_TOPOLOGY);
		if (!Topology.isValidForNew(topology))
			throw new InitializationException("The topology ["+topology+"] is not a valid topology to build new Nodes.");
	}
	
	/**
	 * Initializes a new instance of NodeFactory with the classes
	 * <b>idFactory</b> and <b>nodeClass</b> specifieds by parameter.
	 * If exists any problem with this classes an exception is thrown
	 * with a short description for the problem.
	 * @param idFactory IdFactory class that permits build new Ids for 
	 * the new Nodes.
	 * @param nodeClass Class for new Nodes.
	 * @param topology Topology of the target network.
	 * @throws InitializationException If exists any problem with
	 * initialization of both classes.
	 * @see planet.commonapi.Id Id
	 */
	public NodeFactoryImpl(Class idFactory, Class nodeClass, Class idClass, String topology,int networkSize) throws InitializationException {
		//read Properties.DEFAULT_IDFACTORY
		try {
			//Class idClass,String topology,int networkSize
			Class[] types = {Class.class,String.class,int.class};
			Constructor idFactCons = idFactory.getConstructor(types);
			Object[] params = {idClass,topology,new Integer(networkSize)};
			this.idFactory = (IdFactory)idFactCons.newInstance(params);
		} catch (InstantiationException e) {
			throw new InitializationException(e);
		} catch (IllegalAccessException e) {
			throw new InitializationException(e);
		} catch (Exception e) {
			throw new InitializationException("Property ["+Properties.DEFAULT_IDFACTORY +"] is not a valid class name or some error occur.",e);
		}
		
		//read Properties.DEFAULT_NODECLASS
		try {
			Class types[] = {Id.class};
			this.nodeClass = nodeClass;
			nodeClassConstructor = this.nodeClass.getConstructor(types);
		} catch (SecurityException e) {
			throw new InitializationException("Cannot access to the Constructor "+nodeClass+"(Id)",e);
		} catch (NoSuchMethodException e) {
			throw new InitializationException("Not exists Constructor "+nodeClass+"(Id)",e);
		} catch (Exception e) {
			throw new InitializationException("There is some error during constructor of nodeclass.",e);
		}
		
		this.topology = topology;
		if (!Topology.isValid(topology))
			throw new InitializationException("The topology ["+topology+"] is not a valid topology to build new Nodes.");		
	}
	
	/**
	 * Initializes a new instance of NodeFactory with the 
	 * <b>idFactory</b> and <b>nodeClass</b> specifieds by parameter.
	 * @param idFactory Instance of IdFactory that permits build new Ids for 
	 * the new Nodes.
	 * @param nodeClass Class for new Nodes.
	 * @param topology Topology of the target network.
	 * @throws InitializationException If exists any problem with
	 * initialization.
	 * @see planet.commonapi.Id Id
	 */
	public NodeFactoryImpl(IdFactory idFactory, Class nodeClass, String topology) throws InitializationException {
		this.idFactory = idFactory;
		
		try {
			Class types[] = {Id.class};
			this.nodeClass = nodeClass;
			nodeClassConstructor = this.nodeClass.getConstructor(types);
		} catch (SecurityException e) {
			throw new InitializationException("Cannot access to the Constructor "+nodeClass+"(Id)",e);
		} catch (NoSuchMethodException e) {
			throw new InitializationException("Not exists Constructor "+nodeClass+"(Id)",e);
		}
		
		//test if the topology permits to build new Nodes.
		if (!Topology.isValidForNew(topology))
			throw new InitializationException("The topology ["+topology+"] is not a valid topology to build new Nodes.");
		this.topology = topology;
		
	}
	
	
	/**
	 * Builds a NodeImpl with the actual IdFactory and class for the nodes.
	 * @see planet.commonapi.factory.NodeFactory#buildNode()
	 * @see planet.commonapi.factory.IdFactory IdFactory
	 * @see planet.commonapi.Id Id
	 * @see planet.commonapi.NodeImpl NodeImpl
	 */
	public Node buildNode() throws InitializationException {
		return buildNode(idFactory.buildId());
	}

	/**
	 * Builds a node with the actual implementation class of NodeImpl and the
	 * specified <b>id</b>.
	 * @param id Id to assign to the new instance of NodeImpl.
	 * @return An instance of actual implementation class of NodeImpl.
	 * @see planet.commonapi.factory.NodeFactory#buildNode(planet.commonapi.Id)
	 * @see planet.commonapi.factory.IdFactory IdFactory
	 * @see planet.commonapi.Id Id
	 * @see planet.commonapi.NodeImpl NodeImpl
	 */
	public Node buildNode(Id id) throws InitializationException {
		Node node = null;
		Object params[] = {id};
		try {
			node = (Node) nodeClassConstructor.newInstance(params);
		} catch (IllegalArgumentException e) {
			throw new InitializationException("Bad arguments at Constructor "+nodeClass+"(Id)",e);
		} catch (Exception e) {
			throw new InitializationException("Instantiation problems at Constructor "+nodeClass+"(Id)",e);
		}
		return node;
	}
	
	/**
	 * Gets the actual IdFactory in use.
	 * @see planet.commonapi.factory.NodeFactory#getIdFactory()
	 * @return The actual IdFactory in use.
	 */
	public IdFactory getIdFactory() {
		return idFactory;
	}
	
	/**
	 * Sets the IdFactory to use in the future.
	 * @see planet.commonapi.factory.NodeFactory#setIdFactory(planet.commonapi.IdFactory)
	 * @param idFactory The new IdFactory to use in the future.
	 */
	public void setIdFactory(IdFactory idFactory) {
		this.idFactory = idFactory;
	}
	
	/**
	 * Gets the actual class of implementation of NodeImpl returned by this factory.
	 * @see planet.commonapi.factory.NodeFactory#getNodeClass()
	 * @return The actual class of implementation of NodeImpl returned.
	 */
	public Class getNodeClass() {
		return nodeClass;
	}
	
	/**
	 * Sets the class of an implementation of NodeImpl to return by this
	 * factory in the future.
	 * @see planet.commonapi.factory.NodeFactory#setNodeClass(java.lang.Class)
	 * @param nodeClass Class of an implementation of NodeImpl to return by
	 * this factory.
	 */
	public void setNodeClass(Class nodeClass) throws InitializationException {
		this.nodeClass = nodeClass;
		try {
			Class types[] = {Id.class};
			nodeClassConstructor = this.nodeClass.getConstructor(types);
		} catch (SecurityException e) {
			throw new InitializationException("Cannot access to the Constructor "+nodeClass+"(Id)",e);
		} catch (NoSuchMethodException e) {
			throw new InitializationException("Not exists Constructor "+nodeClass+"(Id)",e);
		}
	}
	
	/**
	 * Gets the actual topology for the network. It is necessary to known 
	 * how generate new Ids for the nodes.
	 * @see planet.commonapi.factory.NodeFactory#getTopology()
	 * @return The topology of the network actually in use.
	 */
	public String getTopology() {
		return topology;
	}
	
	/**
	 * Sets the topology of the network to know how generates new Ids.
	 * @param topology Topology of the network.
	 * @see planet.commonapi.factory.NodeFactory#setTopology(java.lang.String)
	 */
	public void setTopology(String topology) throws InitializationException {
		if (!Topology.isValid(topology))
			throw new InitializationException("Topology ["+topology+"] is not a valid network topology");
		this.topology = topology;
	}
	
	/**
	 * Makes nothing special, only invokes to <b>stream.defaultWriteObject()</b>.
	 * @param stream Stream to save the actual instance.
	 * @throws IOException if any error has ocurred.
	 */
	private void writeObject(java.io.ObjectOutputStream stream)
    throws IOException {
		stream.defaultWriteObject();
	}

	/**
	 * Makes a first <b>stream.defaultReadObject()</b> invocation to uses
	 * the default read method. And then, builds the constructor for
	 * the loaded Node class.
	 * @param stream Stream to read the actual instance.
	 * @throws IOException if occur any error during 
	 * default read object, or if there are any error during
	 * building of Node constructor.  
	 * @throws ClassNotFoundException
	 */
	private void readObject(java.io.ObjectInputStream stream)
    throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		try {
			Class types[] = {Id.class};
			nodeClassConstructor = this.nodeClass.getConstructor(types);
		} catch (SecurityException e) {
			throw new IOException("Cannot access to the Constructor "+nodeClass+"(Id)");
		} catch (NoSuchMethodException e) {
			throw new IOException("Not exists Constructor "+nodeClass+"(Id)");
		}
	}
}
