package planet.symphony;
import planet.util.*;
import planet.util.Properties;

import java.util.*;

import planet.commonapi.*;
import planet.commonapi.exception.InitializationException;
import planet.generic.commonapi.factory.GenericFactory;
import planet.simulate.Globals;
import planet.simulate.Logger;
import planet.simulate.MessageListener;
import planet.simulate.MessagePool;
import planet.simulate.Results;

import planet.symphony.listeners.*;
import planet.symphony.messages.*;

/**
 * 
 * This node is an implementation of the Symphony overlay. 
 * 
 * @author <a href="mailto:heliodoro.tejedor@estudiants.urv.es">Helio Tejedor</a>
 * @author <a href="mailto:Ruben.Mondejar@estudiants.urv.es">Ruben Mondejar</a>
 * @author <a href="mailto:cpairot@etse.urv.es">Carles Pairot</a>
 * @author <a href="mailto:jordi.pujol@estudiants.urv.es">Jordi Pujol</a>
 */

public class SymphonyNode
	extends planet.generic.commonapi.NodeImpl
{
	
	/* ******************  CONSTANTS FOR MODE OF ROUTEMESSAGE *******/
	
	/**
	 * Mode value: Defines the start of communication that requires 
	 * response (REPLY). 
	 */
	public final static int REQUEST					= 0;
	/**
	 * Mode value: Defines the response of a communication.
	 */
	public final static int REPLY					= 1;
	/**
	 * Mode value: Defines a message's mode that requires only
	 * communication on one way.
	 */
	public final static int REFRESH  				= 2;		
	
	/* END ******************  CONSTANTS FOR MODE OF ROUTEMESSAGE *******/
	
	/* ******************  CONSTANTS FOR TYPE OF ROUTEMESSAGE *******/
	/**
	 * Type value: First message that is send by any node to get its
	 * inmediate successor.
	 */
	 public final static int QUERY_JOIN 			= 3;
     public final static int QUERY_JOIN_SET 		= 4;
     public final static int SET_INFO   			= 5;
     public final static int JOIN_FINISHED 			= 6;
     public final static int SET_SECTIONSIZEPRED 	= 7;
     public final static int REFUSE_JOIN 			= 8;
     public final static int QUERY_CONNECT 			= 9;
     public final static int ACCEPT_CONNECT 		= 10;
     public final static int CANCEL_CONNECT 		= 11;
     public final static int CLOSE_LONG_CONNECT 	= 12;
     public final static int CLOSE_PRED_CONNECT 	= 13;
     public final static int CLOSE_SUCC_CONNECT 	= 14;
     public final static int NEW_SUCC_IN_LIST 		= 15;
     public final static int GET_SUCC 				= 16;

	/**
	 * Type value: The message with this type identifies a broadcast message.
	 */
	public final static int BROADCAST  				= 17;
	/**
	 * Type value: It identifies that this message contains an application 
	 * level Message. 
	 */
	public final static int DATA 					= 18;	
	
	/* END ******************  CONSTANTS FOR TYPE OF ROUTEMESSAGE *******/
	
	/* ******************  CONSTANTS FOR TYPE/MODE OF ROUTEMESSAGE *******/
	/**
	 * This String contains a string representation of each of types
	 * and modes values of the RouteMessage.
	 */
	public final static String[] TYPES = {"REQUEST","REPLY","REFRESH",
			"QUERY_JOIN", "QUERY_JOIN_SET","SET_INFO", "JOIN_FINISHED", 
			"SET_SECTIONSIZEPRED","REFUSE_JOIN", "QUERY_CONNECT", 
			"ACCEPT_CONNECT", "CANCEL_CONNECT", "CLOSE_LONG_CONNECT", 
			"CLOSE_PRED_CONNECT","CLOSE_SUCC_CONNECT",
			"NEW_SUCC_IN_LIST","GET_SUCC","BROADCAST","DATA"}; 
	
	/* END ******************  CONSTANTS FOR TYPE/MODE OF ROUTEMESSAGE *******/
	
	
	
	/**
	 * To build new Id to find a new long distance.
	 */
	private static Random r = new Random();

	
	private Set incommingSet;
	private Set outcommingSet;
	private TreeSet succSet, predSet;  // with order !!!
	
	private double n;
	
	/* UpdateOutcommingList */
	private Set newSet;
	private Set connectedNodes;
	private int retriesNewLongDistance;
	
	/* QueryJoin */
	private NodeHandle curNew;
	private Set pendingJoinMsg;
		
	private static final int CAN_ACCEPT_JOIN = 0;
	private static final int JOINING_ME = 1;
	private static final int WAIT_JOIN_FINISHED = 2;
	private static final String[] queryJoinStateString = { "CAN_ACCEPT_JOIN", "JOINING_ME", "WAIT_JOIN_FINISHED" };	
	private int queryJoinState;
	
	private boolean alive;
	private boolean stabilized;

	public SymphonyNode (Id id) throws InitializationException {
		super (id);
		endpoints = new Hashtable();
		alive = true;
		
		outcommingSet = new HashSet();
		incommingSet = new HashSet();
		succSet = new TreeSet(new SuccSetComparator(id));
		predSet = new TreeSet(new PredSetComparator(id));
		
		newSet = new HashSet();
		connectedNodes = new HashSet();
		pendingJoinMsg = new HashSet();
		stabilized = true;
	}
  
    private String sendMessage(NodeHandle from, NodeHandle to, int type, int mode, Message m) {
		String key = GenericFactory.generateKey();
		NodeHandle nextHop = route(to);
		return sendMessage(from, to, route(to), type, mode, m, key);
	}
	
	private String sendMessage(NodeHandle from, NodeHandle to, int type, int mode, Message m, String key) {
		NodeHandle nextHop = route(to);
		return sendMessage(from, to, nextHop, type, mode, m, key);
	}
	
	private String sendMessage(NodeHandle from, NodeHandle to, NodeHandle nextHop, int type, int mode, Message m) {
		String key = GenericFactory.generateKey();
		return sendMessage(from, to, nextHop, type, mode, m, key);
	}
	
	private String sendMessage(NodeHandle from, NodeHandle to, NodeHandle nextHop, int type, int mode, Message m, String key) {
		RouteMessage msg = null;
		try {
			msg = MessagePool.getMessage(key, from, to, type, mode);
			msg.setNextHopHandle(nextHop);
			msg.setMessage(m);
			send(msg);
		} catch (QueueFull e) {
			key = null;
			Logger.log("Outgoing Queue of Node " + id + " is Full", Logger.ERROR_LOG);
			MessagePool.freeMessage(msg);
		} catch (InitializationException e) {
			key = null;
			Logger.log("ERROR: Cannot get a RouteMessage of MessagePool\n" + e.getMessage(), Logger.ERROR_LOG);
		}
		return key;
	}
	
	private String resendMessage(RouteMessage msg) {
		return resendMessage(msg, msg.getNextHopHandle());
	}
	
	private String resendMessage(RouteMessage msg, NodeHandle nextHop) {
		String key = null;
		try {
			key = msg.getKey();
			msg.setNextHopHandle(nextHop);
			send(msg);
		} catch (QueueFull e) {
			key = null;
			Logger.log("Outgoing Queue of Node " + id + " is Full", Logger.ERROR_LOG);
			MessagePool.freeMessage(msg);
		}
		return key;
	}
	
	private boolean addToSuccSet(NodeHandle newSucc) {
		 boolean inserted;
		
		if (succSet.contains(newSucc) || (succSet.size() > 0 && newSucc.getId().equals(id))) {
			inserted = false;
		} else {
			if (succSet.size() > 0 && !newSucc.getId().equals(id))
				succSet.remove(nodeHandle);
			succSet.add(newSucc);
			while (succSet.size() > ((SymphonyProperties)Properties.overlayProps).F) {
				NodeHandle tokill = (NodeHandle) succSet.last();
				succSet.remove(tokill);
				sendMessage(nodeHandle, tokill, tokill, CLOSE_PRED_CONNECT, REFRESH, null);
			}
			inserted = succSet.contains(newSucc);
		}

		if (inserted) {
			updateOutcommingList();
			SetInfoMessage sim = new SetInfoMessage();
			Collection c = new HashSet();
			c.add(nodeHandle);
			sim.setPredSet(c);
			sendMessage(nodeHandle, newSucc, newSucc, SET_INFO, REFRESH, sim);
		}
		return inserted;
	}
	
	private boolean addToPredSet(NodeHandle newPred) {
		boolean inserted;
		if (!predSet.contains(newPred)) {
			predSet.add(newPred);
			if (predSet.size() > 1)
				predSet.remove(nodeHandle);
			inserted = true;
		} else {
			inserted = false;
		}
		
		if (inserted) {
			updateOutcommingList();
		}
		return inserted;
	}
	
	private boolean addToSuccSet(Collection c) {
		Iterator it = c.iterator();
		boolean ret = false;
		while (it.hasNext()) {
			ret |= addToSuccSet( (NodeHandle) it.next() );
		}
		return ret;
	}
	
	private boolean addToPredSet(Collection c) {
		Iterator it = c.iterator();
		boolean ret = false;
		while (it.hasNext()) {
			ret |= addToPredSet( (NodeHandle) it.next() );
		}
		return ret;
	}
	
	private NodeHandle getSucc() {
		if (succSet.isEmpty())
			return nodeHandle;
		else
			return (NodeHandle) succSet.first();
	}
	
	public NodeHandle getPred() {
		if (predSet.isEmpty())
			return nodeHandle;
		else
			return (NodeHandle) predSet.first();
	}
	
  	public void join(Id bsId) {
  		succSet.clear();
		predSet.clear();
		incommingSet.clear();
		outcommingSet.clear();
		newSet.clear();
		connectedNodes.clear();
		pendingJoinMsg.clear();
		
		if (id.equals(bsId)) {
			n = 1;
			queryJoinState = CAN_ACCEPT_JOIN;
			stabilized = true;
		} else {
			n = 0;
			try {
				NodeHandle bootstrap = GenericFactory.buildNodeHandle(bsId,true);
				sendMessage(nodeHandle, nodeHandle, bootstrap, QUERY_JOIN, REFRESH, new JoinMessage(nodeHandle, bootstrap));
			} catch (InitializationException e) {
				Logger.log("["+id+"] Cannot to build a NodeHandle before sending a RouteMessage",Logger.ERROR_LOG);
			}
			queryJoinState = JOINING_ME;
			stabilized = false;
		}
	}
	
	public void leave() {
		Iterator[] c = { 
				incommingSet.iterator(),
				outcommingSet.iterator(),
				succSet.iterator(),
				predSet.iterator()
			};
			int[] msgtype = {
				CLOSE_LONG_CONNECT,
				CLOSE_LONG_CONNECT,
				CLOSE_PRED_CONNECT,
				CLOSE_SUCC_CONNECT
			};
			for(int i = 0; i < c.length; i++) {
				Iterator it = c[i];
				while (it.hasNext()) {
					NodeHandle current = (NodeHandle) it.next();
					sendMessage(nodeHandle, current, current, msgtype[i], REFRESH, null);
				}
			}

			stabilized = true;
			alive = false;
	}

	public void lookup(Id key) {
		String msgkey = null;;
		try {
			msgkey = sendMessage(nodeHandle, GenericFactory.buildNodeHandle(key,true), Globals.LOOKUP, REQUEST, null);
		} catch (InitializationException e) {
			Logger.log("["+id+"] cannot build a NodeHandle for the key ["+key+"]",Logger.ERROR_LOG);
		}
		addMessageListener(msgkey, new LookupListener(id, key));
		Results.incTraffic();
	}

	private void demandNewLongDistance() {
		int retries = 10;
		boolean sortir = false;
		
		while (!sortir) {
			SymphonyId symId = (SymphonyId) id;
			double x = symId.getDoubleValue() + Math.exp(Math.log((double) n) * (r.nextDouble() - 1.0));
			int i = (int) x;
			x -= i;
			Id xid = new SymphonyId(x);
			NodeHandle succ = getSucc();
			NodeHandle pred = getPred();
			if (!xid.between(pred.getId(), id) && !xid.between(id, succ.getId())) {
				try {
					sendMessage(nodeHandle, GenericFactory.buildNodeHandle(xid,true), nodeHandle, QUERY_CONNECT, REFRESH, null);
				} catch (InitializationException e) {
					Logger.log("["+id+"] cannot build a NodeHandle for a RouteMessage",Logger.ERROR_LOG);
				}
				sortir = true;
			} else {
				if (retries > 1) 
					retries--;
				else {
					Logger.log(id + ": Cannot find a valid Id to connect it!!!!!", Logger.MSG_LOG);
					stabilized = true;
					sortir = true;
				}
			}
		}
	}

	private void updateOutcommingList() {
		double nnew = estimation();
	
		boolean recalc = true;
		if ( (outcommingSet.size() > 0) && (n != 0) ) {
			double ratio = (double) nnew / (double) n;
			if ( (ratio >= 0.5) && (ratio <= 2) ) {
				recalc = false;
			}
		}

		if (recalc) {
			n = nnew;
			retriesNewLongDistance = 0;
			stabilized = false;
			demandNewLongDistance();
		}
	}
	
	private double estimation() {
		double sum = 0.0;
		int cnt = 0;

		Iterator it = predSet.iterator();
		Id to = id;
		Id from;
		while (cnt < 2 && it.hasNext()) {
			from = ((NodeHandle) it.next()).getId();
			sum += ((SymphonyId) to.subtract(from)).getDoubleValue();
			cnt++;
			to = from;			
		}

		if (!succSet.isEmpty()) {
			from = id;
			to = ((NodeHandle) succSet.first()).getId();
			sum = ((SymphonyId) to.subtract(from)).getDoubleValue();
			cnt++;
		}
		
		if (cnt == 0)
			return 1.0;
		else
			return (double) cnt / sum;
	}

	public void dispatcher(RouteMessage msg) {
		/* Broadcast still unavailable */
		NodeHandle src = msg.getSource();
		NodeHandle dest = msg.getDestination();
		
		if (!alive) {
			Logger.log("Node " + id + " is not in, but it receives messages!!", Logger.ERROR_LOG);
			MessagePool.freeMessage(msg);
			return;
		}
		
		if (dest == null) {
			Logger.log("destination null", Logger.ERROR_LOG);
			MessagePool.freeMessage(msg);
			return;
		}
		
		if (!dest.equals(nodeHandle) && getPred() != null && !dest.getId().between(getPred().getId(), id)) {
			if (msg.getType() == Globals.LOOKUP && msg.getMode() == REQUEST) {
				Results.updateHopsMsg(id,msg.getKey());
			}
			
			NodeHandle proposed = route(dest);
			msg.setNextHopHandle(proposed);
			try {
				send(msg);
			} catch (QueueFull e) {
				Logger.log("Outgoing Queue of Node " + this.id + " is Full", Logger.ERROR_LOG);
				MessagePool.freeMessage(msg);
			}
		} else {
			switch(msg.getMode()) {
			// Mode = REQUEST
			case REQUEST: {
				switch(msg.getType()) {
				// Mode = REQUEST, Type = LOOKUP
				case Globals.LOOKUP: {
					sendMessage(nodeHandle, msg.getSource(), Globals.LOOKUP, REPLY, msg.getMessage(), msg.getKey());
					break;
				}
				
				// Mode = REQUEST, Type Error
				default:
					Logger.log("Uncatched message Type [" + msg.getType() + "]", Logger.ERROR_LOG);
				}
				break;
			}
			
			// Mode = REPLY
			case REPLY: {
				String key = msg.getKey();
				try {
					MessageListener ml = (MessageListener) listeners.get(msg.getKey());
					ml.onMessage(msg);
				} catch (NullPointerException e) {
					Logger.log("I'm [" + id + "]; not exist listener for key [" + msg.getKey() + "]", Logger.ERROR_LOG);
				}
				removeMessageListener(key);
				break;
			}
			
			// Mode = REFRESH
			case REFRESH: {
				switch(msg.getType()) {
				// Mode = REFRESH, Type = QUERY_JOIN
				case QUERY_JOIN: {
					switch(queryJoinState) {
					case CAN_ACCEPT_JOIN:
						sendMessage(nodeHandle, getPred(), Globals.JOIN, REFRESH, msg.getMessage());
						curNew = msg.getSource();
						queryJoinState = WAIT_JOIN_FINISHED;
						break;
					case JOINING_ME:
						sendMessage(nodeHandle, msg.getSource(), msg.getSource(), REFUSE_JOIN, REFRESH, msg.getMessage());
						break;						
					case WAIT_JOIN_FINISHED:
						pendingJoinMsg.add(msg);
						Logger.log("WAIT_JOIN_FINISHED enqueue msg from " + msg.getSource(), Logger.MSG_LOG);
						break;
					}
					break;
				}
				
				// Mode = REFRESH, Type = JOIN
				case Globals.JOIN: {
					JoinMessage jm = (JoinMessage) msg.getMessage();
					NodeHandle newNode = jm.getNewNode();

					NodeHandle succSucc = getSucc();
					NodeHandle succ = newNode;
					NodeHandle pred = getPred();
					
					TreeSet ss = (TreeSet) succSet.clone();
					ss.add(nodeHandle);
					if (ss.size() > ((SymphonyProperties)Properties.overlayProps).F) {
						ss.remove(ss.last());
					}
					
					Collection ps = (Collection) predSet.clone();
					ps.add(nodeHandle);
					
					SetInfoMessage sim = new SetInfoMessage();
					sim.setSuccSet(ss);
					sim.setPredSet(ps);
					sendMessage(nodeHandle, newNode, newNode, SET_INFO, REFRESH, sim);

					NewSuccInListMessage nsilm = new NewSuccInListMessage(newNode, ((SymphonyProperties)Properties.overlayProps).F);
					sendMessage(nodeHandle, pred, NEW_SUCC_IN_LIST, REFRESH, nsilm);
					
					addToSuccSet(newNode);
					
					break;
				}
				
				// Mode = REFRESH, Type = SET_INFO
				case SET_INFO: {
					SetInfoMessage sim = (SetInfoMessage) msg.getMessage();
					if (sim.hasSuccSet()) addToSuccSet(sim.getSuccSet());
					if (sim.hasPredSet()) addToPredSet(sim.getPredSet());
					if (queryJoinState == JOINING_ME) {
						queryJoinState = CAN_ACCEPT_JOIN;
						sendMessage(nodeHandle, getSucc(), JOIN_FINISHED, REFRESH, null);
					}
					break;
				}

				// Mode = REFRESH, Type = JOIN_FINISHED
				case JOIN_FINISHED: {
					NodeHandle newNode = msg.getSource();
					NodeHandle oldPred = getPred();
					
					Iterator it = pendingJoinMsg.iterator();
					while (it.hasNext()) {
						RouteMessage newMsg = (RouteMessage) it.next();
						NodeHandle newMsgDest;
						if (newMsg.getSource().getId().between(newNode.getId(), id))
							newMsgDest = newNode;
						else
							newMsgDest = oldPred;
						try {
							newMsg.setNextHopHandle(newMsgDest);
							send(newMsg);
						} catch (QueueFull e) {
							Logger.log("Outgoing Queue of Node " + id + " is Full", Logger.ERROR_LOG);
							MessagePool.freeMessage(msg);
						}
					}
					queryJoinState = CAN_ACCEPT_JOIN;
					pendingJoinMsg.clear();
					
					break;
				}
				// Mode = REFRESH, Type = REFUSE_JOIN
				case REFUSE_JOIN: {
					JoinMessage jm = (JoinMessage) msg.getMessage();
					NodeHandle bsHandle = jm.getBootStrap();
					int retries = jm.getRetries();
					
					if (retries < ((SymphonyProperties)Properties.overlayProps).maxJoinRetries) {
						jm.setRetries(retries + 1);
						sendMessage(nodeHandle, nodeHandle, bsHandle, QUERY_JOIN, REFRESH, jm);
					} else {
						Logger.log("Node " + id + " has wasted the retries to join to the network", Logger.ERROR_LOG);
						stabilized = true;
					}
					break;
				}
				
				// Mode = REFRESH, Type = QUERY_CONNECT
				case QUERY_CONNECT: {
					NodeHandle srcHandle = msg.getSource();
					String key = msg.getKey();
					
					boolean accepted;
					if (srcHandle.getId().equals(id) 
							|| predSet.contains(srcHandle) 
							|| succSet.contains(srcHandle)
							|| newSet.contains(srcHandle)
							|| outcommingSet.contains(srcHandle)) {
						accepted = false;
					} else if (incommingSet.contains(srcHandle)) {
						accepted = true;
					} else if (incommingSet.size() < ((SymphonyProperties)Properties.overlayProps).K) {
						incommingSet.add(srcHandle);
						accepted = true;
					} else {
						accepted = false;
					}
					
					msg.setType( accepted ? ACCEPT_CONNECT : CANCEL_CONNECT );
					msg.setDestination(msg.getSource());
					msg.setSource(nodeHandle);
					try {
						send(msg);
					} catch (QueueFull e) {
						Logger.log("Outgoing Queue of Node " + id + " is Full", Logger.ERROR_LOG);
						MessagePool.freeMessage(msg);
					}
					break;
				}
				
				// Mode = REFRESH, Type = ACCEPT_CONNECT
				case ACCEPT_CONNECT: {
					NodeHandle longHandle = msg.getSource();
					if (incommingSet.contains(longHandle)) {
						if (id.compareTo(longHandle.getId()) < 0) {
							incommingSet.remove(longHandle);
						} else {
							demandNewLongDistance();
							break;
						}
					}
					boolean yaExiste = newSet.contains(longHandle);
					if (!predSet.contains(longHandle) && !succSet.contains(longHandle) && !longHandle.equals(nodeHandle)) {
						newSet.add(longHandle);
						outcommingSet.add(longHandle); 
						incommingSet.remove(longHandle); 
						connectedNodes.remove(longHandle);
					}
					
					if (!yaExiste)
						retriesNewLongDistance = 0;
					
					if (newSet.size() == ((SymphonyProperties)Properties.overlayProps).K) {
						Set tmp = outcommingSet;
						outcommingSet = newSet;
						newSet = tmp;
						newSet.clear();
						
						Iterator it = connectedNodes.iterator();
						while (it.hasNext()) {
							NodeHandle current = (NodeHandle) it.next();
							sendMessage(nodeHandle, current, CLOSE_LONG_CONNECT, REFRESH, null);
						}
						connectedNodes.clear();
						stabilized = true;
					} else {
						demandNewLongDistance();
					}
					break;
				}
				
				// Mode = REFRESH, Type = CANCEL_CONNECT
				case CANCEL_CONNECT: {
					retriesNewLongDistance++;
					if (retriesNewLongDistance < ((SymphonyProperties)Properties.overlayProps).maxRetriesNewLongDistance) 
						demandNewLongDistance();
					else
						stabilized = true;
					break;
				}
				
				// Mode = REFRESH, Type = CLOSE_CONNECT
				case CLOSE_LONG_CONNECT: {
					NodeHandle srcHandle = msg.getSource();
					
					newSet.remove(srcHandle);
					connectedNodes.remove(srcHandle);
					incommingSet.remove(srcHandle);
					boolean outremoved = outcommingSet.remove(srcHandle);
					if (outremoved) {
						if (stabilized)
							demandNewLongDistance();
						else
							newSet.remove(srcHandle);
					}
					break;
				}
				//Mode = REFRESH, Type = CLOSE_PRED_CONNECT
				case CLOSE_PRED_CONNECT: {
					NodeHandle srcHandle = msg.getSource();
					predSet.remove(srcHandle);
					break;
				}
				
				// Mode = REFRESH, Type = CLOSE_SUCC_CONNECT
				case CLOSE_SUCC_CONNECT: {
					NodeHandle srcHandle = msg.getSource();
					
					if (succSet.remove(srcHandle))
						if (!succSet.isEmpty())
							sendMessage(nodeHandle, (NodeHandle) succSet.last(), GET_SUCC, REFRESH, null);
					break;
				}
				
				// Mode = REFRESH, Type = GET_SUCC
				case GET_SUCC: {
					SetInfoMessage sim = new SetInfoMessage();
					Collection c = new HashSet();
					c.add(getSucc());
					sim.setSuccSet(c);
					sendMessage(nodeHandle, msg.getSource(), SET_INFO, REFRESH, sim);
					break;
				}
				
				// Mode = REFRESH, Type = NEW_SUCC_IN_LIST
				case NEW_SUCC_IN_LIST: {
					NewSuccInListMessage nsilm = (NewSuccInListMessage) msg.getMessage();
					NodeHandle newNode = nsilm.getNewNode();
					addToSuccSet(newNode);
					if (nsilm.decF() >= 0) {
						sendMessage(nodeHandle, getPred(), NEW_SUCC_IN_LIST, REFRESH, nsilm);
					}
					break;
				}
				
				// Mode = REFRESH, Type Error
				default:
					Logger.log("Uncatched message Type [" + msg.getType()+ "]", Logger.ERROR_LOG);
				}
				break;
			}
			
			// Mode = ERROR
			case Globals.ERROR:
				switch(msg.getType()) {
					case QUERY_CONNECT:
						demandNewLongDistance();
						break;
					case ACCEPT_CONNECT:
						NodeHandle srcHandle = msg.getSource();
						incommingSet.remove(srcHandle);
						break;
					default:
						Logger.log("Uncatched message Mode [" + msg.getMode() + "]", Logger.ERROR_LOG);
					}
				break;
			
			default:
				Logger.log("Uncatched message Mode [" + msg.getMode() + "]", Logger.ERROR_LOG);
			}
		}
	}

	public boolean process(int actualStep) {
		super.process(actualStep);
		while (hasMoreMessages()) {      
			dispatcher(nextMessage());
		}

		/* APPEAR IMPLEMENTED IN THE FUTURE
		 * if (timeKeepAlive > Properties.keepAliveSteps) {
			exchange_some_information_piggy_backed_on_keep-alives;
			timeKeepAlive = 0;
		} else {
			timeKeepAlive++;
		}*/
		return !IsStabilized(); //true if to continue
	}

	public Hashtable getInfo() {
		Hashtable info = new Hashtable();
		info.put("predSet", predSet.clone());
		info.put("succSet", succSet.clone());
		info.put("estimation", new Double(n));
		return info;
	}
	
	
	
	/**
	 * @param appId String that includes the application identification.
	 * @param to Node that must receive this message. If cannot represents
	 * a real node.
	 * @param nextHop Node that must receive this message as first hop.
	 * @param msg Message to be sent.
	 * If it is null, the message must be routed.
	 */
	public void routeData(String appId,
			NodeHandle to, NodeHandle nextHop, Message msg) {
		NodeHandle aux = null;
		if (nextHop == null) {
			aux = route(to);
			if (nextHop == null) {
				Logger.log("route return null!", Logger.ERROR_LOG);
			}
		}
		
		if (aux != null) {
			try {
				sendData(MessagePool.getMessage("",nodeHandle,to,aux,msg,DATA,REQUEST,appId), nextHop);
			} catch (InitializationException e) {
				Logger.log("["+id+"] cannot build a RouteMessage.",Logger.ERROR_LOG);
			}
		}
	}
	
	public void printNode() {
		StringBuffer sb = new StringBuffer();
		sb.append(
			queryJoinStateString[queryJoinState] +
			";" + id + 
			";" + n +
			";"
		);
		Iterator it;
		
		sb.append("OutcommingList;");
		it = outcommingSet.iterator();
		while (it.hasNext()) {
			NodeHandle current = (NodeHandle) it.next();
			sb.append(current.toString() + ";");
		}
		
		sb.append("Incomming;");
		it = incommingSet.iterator();
		while (it.hasNext()) {
			NodeHandle current = (NodeHandle) it.next();
			sb.append(current.toString() + ";");
		}
		sb.append("succSet;");
		it = succSet.iterator();
		while (it.hasNext()) {
			NodeHandle current = (NodeHandle) it.next();
			sb.append(current.getId().toString() + ";");
		}
		sb.append("predSet;");
		it = predSet.iterator();
		while (it.hasNext()) {
			NodeHandle current = (NodeHandle) it.next();
			sb.append(current.toString() + ";");
		}
		sb.append("Fin");
		
		System.out.println(sb.toString());
	}
	
	public void prettyPrintNode() {
	}
	
	
	public boolean isAlive() {
		return alive;
	}
	
	/* Privates */
	
	private NodeHandle bestOf(NodeHandle to, NodeHandle a, NodeHandle b) {
		if (a == null)
			return b;			
		if (b == null)
			return a;
		double lena = ((SymphonyId) to.getId().subtract(a.getId())).getDoubleValue();
		double lenb = ((SymphonyId) to.getId().subtract(b.getId())).getDoubleValue();
		if (lena <= lenb)
			return a;
		else
			return b;
	}

	private NodeHandle bestOf(NodeHandle to, NodeHandle a, Iterator it) {
		NodeHandle candidate = a;
		
		while (it.hasNext()) {
			NodeHandle current = (NodeHandle) it.next();
			candidate = bestOf(to, candidate, current);
		}
		return candidate;
	}
	
	private NodeHandle route(NodeHandle to) {
		if (to.getId().between(id, getSucc().getId()))
			return getSucc();

		NodeHandle candidate;
		Iterator it;
		
		candidate = null;
		
		it = succSet.iterator();
		candidate = bestOf(to, candidate, it);
		
		it = predSet.iterator();
		candidate = bestOf(to, candidate, it);
		
		it = outcommingSet.iterator();
		candidate = bestOf(to, candidate, it);
		
		/* bidirectional ? */
		it = incommingSet.iterator();
		candidate = bestOf(to, candidate, it);
		
		/* optimize? */
		it = newSet.iterator();
		candidate = bestOf(to, candidate, it);
		
		it = connectedNodes.iterator();
		candidate = bestOf(to, candidate, it);
			
		return candidate;
	}
	
	private void sendData(RouteMessage message, NodeHandle handle) {
		message.setType(DATA);
		message.setMode(REFRESH);
		message.setNextHopHandle(handle);

		try {
			send(message);
			Results.incTraffic();
		} catch (QueueFull e) {
			Logger.log("Outgoing Queue of Node "+ id + " is Full", Logger.ERROR_LOG);
			MessagePool.freeMessage (message);
		}
	}

	/* misc */
	
	public String toString() {
		return "SymphonyNode(" + id + ")";
	}
	
	public boolean IsStabilized() {
		return stabilized;
	}
	
	/* ******** TO IMPLEMENT IN THE FUTURE ********/
	/**
	 * @see planet.commonapi.Node#broadcast(java.lang.String, planet.commonapi.NodeHandle, planet.commonapi.NodeHandle, planet.commonapi.Message)
	 * @param appId
	 * @param to
	 * @param nextHop
	 * @param msg
	 */
	public void broadcast(String appId, NodeHandle to, NodeHandle nextHop,
			Message msg) {
		// TODO Auto-generated method stub
	}
	
	/**
	 * @see planet.commonapi.Node#fail()
	 * 
	 */
	public void fail() {
		// TODO Auto-generated method stub
	}
	
	/**
	 * @see planet.commonapi.Node#getSuccList(int)
	 * @param max
	 * @return
	 */
	public Vector getSuccList(int max) {
		// TODO Auto-generated method stub
		return null;
	}
	/**
	 * @see planet.commonapi.Node#localLookup(planet.commonapi.Id, int, boolean)
	 * @param key
	 * @param max
	 * @param safe
	 * @return
	 */
	public Vector localLookup(Id key, int max, boolean safe) {
		// TODO Auto-generated method stub
		return null;
	}
	/**
	 * @see planet.commonapi.Node#neighborSet(int)
	 * @param max
	 * @return
	 */
	public Vector neighborSet(int max) {
		// TODO Auto-generated method stub
		return null;
	}
	/**
	 * @see planet.commonapi.Node#range(planet.commonapi.NodeHandle, planet.commonapi.Id, planet.commonapi.Id, planet.commonapi.Id)
	 * @param node
	 * @param rank
	 * @param leftKey
	 * @param rightKey
	 * @return
	 */
	public boolean range(NodeHandle node, Id rank, Id leftKey, Id rightKey) {
		// TODO Auto-generated method stub
		return false;
	}
	/**
	 * @see planet.commonapi.Node#replicaSet(planet.commonapi.Id, int)
	 * @param key
	 * @param maxRank
	 * @return
	 */
	public Vector replicaSet(Id key, int maxRank) {
		// TODO Auto-generated method stub
		return null;
	}
}
